from pathlib import Path
from typing import Callable, Dict, List, Tuple

import dash
import pandas as pd
from dash import html
from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.webviz_store import webvizstore

from ..._datainput.fmu_input import scratch_ensemble
from ..._providers import (
    EnsembleSummaryProvider,
    EnsembleSummaryProviderFactory,
    Frequency,
)
from ._callbacks import plugin_callbacks
from ._ensemble_group_tree_data import EnsembleGroupTreeData
from ._layout import main_layout


class GroupTree(WebvizPluginABC):
    """This plugin vizualizes the network tree and displays pressures,
    rates and other network related information.

    ---
    * **`ensembles`:** Which ensembles in `shared_settings` to include.
    * **`gruptree_file`:** `.csv` with gruptree information.
    * **`time_index`:** Frequency for the data sampling.
    ---

    **Summary data**

    This plugin needs the following summary vectors to be exported:
    * FOPR, FWPR, FOPR, FWIR and FGIR
    * GPR for all group nodes in the network
    * GOPR, GWPR and GGPR for all group nodes in the production network \
    (GOPRNB etc for BRANPROP trees)
    * GGIR and/or GWIR for all group nodes in the injection network
    * WSTAT, WTHP, WBHP, WMCTL for all wells
    * WOPR, WWPR, WGPR for all producers
    * WWIR and/or WGIR for all injectors

    **GRUPTREE input**

    `gruptree_file` is a path to a file stored per realization (e.g. in \
    `share/results/tables/gruptree.csv"`).

    The `gruptree_file` file can be dumped to disk per realization by the `ECL2CSV` forward
    model with subcommand `gruptree`. The forward model uses `ecl2df` to export a table
    representation of the Eclipse network:
    [Link to ecl2csv gruptree documentation.](https://equinor.github.io/ecl2df/usage/gruptree.html).

    **time_index**

    This is the sampling interval of the summary data. It is `yearly` by default, but can be set
    to `monthly` if needed.
    """

    def __init__(
        self,
        app: dash.Dash,
        webviz_settings: WebvizSettings,
        ensembles: list,
        gruptree_file: str = "share/results/tables/gruptree.csv",
        rel_file_pattern: str = "share/results/unsmry/*.arrow",
        time_index: str = "yearly",
    ):
        super().__init__()
        assert time_index in [
            "monthly",
            "yearly",
        ], "time_index must be monthly or yearly"
        self._ensembles = ensembles
        self._gruptree_file = gruptree_file

        if ensembles is None:
            raise ValueError('Incorrect argument, must provide "ensembles"')

        sampling = Frequency(time_index)

        self._ensemble_paths: Dict[str, Path] = {
            ensemble_name: webviz_settings.shared_settings["scratch_ensembles"][
                ensemble_name
            ]
            for ensemble_name in ensembles
        }

        provider_factory = EnsembleSummaryProviderFactory.instance()

        self._group_tree_data: Dict[str, EnsembleGroupTreeData] = {}

        sampling = Frequency(time_index)
        for ens_name, ens_path in self._ensemble_paths.items():
            provider: EnsembleSummaryProvider = (
                provider_factory.create_from_arrow_unsmry_presampled(
                    str(ens_path), rel_file_pattern, sampling
                )
            )
            gruptree = read_ensemble_gruptree(ens_name, ens_path, gruptree_file)
            self._group_tree_data[ens_name] = EnsembleGroupTreeData(provider, gruptree)

        self.set_callbacks(app)

    def add_webvizstore(self) -> List[Tuple[Callable, List[Dict]]]:
        functions: List[Tuple[Callable, List[Dict]]] = []
        functions.append(
            (
                read_ensemble_gruptree,
                [
                    {
                        "ens_name": ens_name,
                        "ens_path": ens_path,
                        "gruptree_file": self._gruptree_file,
                    }
                    for ens_name, ens_path in self._ensemble_paths.items()
                ],
            )
        )
        return functions

    @property
    def tour_steps(self) -> List[dict]:
        return [
            {
                "id": self.uuid("layout"),
                "content": "Dashboard vizualizing Eclipse network tree.",
            },
            {
                "id": self.uuid("selections_layout"),
                "content": "Menu for selecting ensemble and other options.",
            },
            {
                "id": self.uuid("filters_layout"),
                "content": "Menu for filtering options.",
            },
            {
                "id": self.uuid("grouptree_wrapper"),
                "content": "Vizualisation of network tree.",
            },
        ]

    @property
    def layout(self) -> html.Div:
        return html.Div(
            children=[
                main_layout(get_uuid=self.uuid, ensembles=self._ensembles),
            ],
        )

    def set_callbacks(self, app: dash.Dash) -> None:
        plugin_callbacks(
            app=app, get_uuid=self.uuid, group_tree_data=self._group_tree_data
        )


@webvizstore
def read_ensemble_gruptree(
    ens_name: str, ens_path: str, gruptree_file: str
) -> pd.DataFrame:
    """Reads the gruptree files for an ensemble from the scratch disk. These
    files can be exported in the FMU workflow using the ECL2CSV
    forward model with subcommand gruptree.

    If BRANPROP is found in the KEYWORD column, then GRUPTREE rows
    are filtered out.

    If the trees are equal in every realization, only one realization is kept.
    """

    ens = scratch_ensemble(ens_name, ens_path, filter_file="OK")
    df_files = ens.find_files(gruptree_file)

    if df_files.empty:
        raise ValueError(f"No gruptree file available for ensemble: {ens_name}")

    # Load all gruptree dataframes and check if they are equal
    compare_columns = ["DATE", "CHILD", "KEYWORD", "PARENT"]
    df_prev = pd.DataFrame()
    dataframes = []
    gruptrees_are_equal = True
    for i, row in df_files.iterrows():
        df_real = pd.read_csv(row["FULLPATH"])

        if "BRANPROP" in df_real["KEYWORD"].unique():
            df_real = df_real[df_real["KEYWORD"] != "GRUPTREE"]
        if (
            i > 0
            and gruptrees_are_equal
            and not df_real[compare_columns].equals(df_prev)
        ):
            gruptrees_are_equal = False
        else:
            df_prev = df_real[compare_columns].copy()

        df_real["REAL"] = row["REAL"]
        dataframes.append(df_real)
    df = pd.concat(dataframes)

    # Return either one or all realization in a common dataframe
    if gruptrees_are_equal:
        df = df[df["REAL"] == df["REAL"].min()]

    df["DATE"] = pd.to_datetime(df["DATE"])

    return df.where(pd.notnull(df), None)



from typing import Any, Callable, Dict, List, Tuple, Union

import webviz_core_components as wcc
from dash import Dash, Input, Output, State
from dash.exceptions import PreventUpdate

from .._business_logic import RftPlotterDataModel, filter_frame
from .._figures._crossplot_figure import update_crossplot
from .._figures._errorplot_figure import update_errorplot
from .._figures._formation_figure import FormationFigure
from .._figures._map_figure import MapFigure
from .._figures._misfit_figure import update_misfit_plot
from .._layout import LayoutElements


def plugin_callbacks(
    app: Dash, get_uuid: Callable, datamodel: RftPlotterDataModel
) -> None:
    @app.callback(
        Output(get_uuid(LayoutElements.FORMATIONS_WELL), "value"),
        Input(get_uuid(LayoutElements.MAP_GRAPH), "clickData"),
    )
    def _get_clicked_well(click_data: Dict[str, List[Dict[str, Any]]]) -> str:
        if not click_data:
            return datamodel.well_names[0]
        for layer in click_data["points"]:
            try:
                return layer["customdata"]
            except KeyError:
                pass
        raise PreventUpdate

    @app.callback(
        Output(get_uuid(LayoutElements.MAP), "children"),
        Input(get_uuid(LayoutElements.MAP_ENSEMBLE), "value"),
        Input(get_uuid(LayoutElements.MAP_SIZE_BY), "value"),
        Input(get_uuid(LayoutElements.MAP_COLOR_BY), "value"),
        Input(get_uuid(LayoutElements.MAP_DATE_RANGE), "value"),
    )
    def _update_map(
        ensemble: str, sizeby: str, colorby: str, dates: List[float]
    ) -> Union[str, List[wcc.Graph]]:

        figure = MapFigure(datamodel.ertdatadf, ensemble)
        if datamodel.faultlinesdf is not None:
            figure.add_fault_lines(datamodel.faultlinesdf)
        figure.add_misfit_plot(sizeby, colorby, dates)

        return [
            wcc.Graph(
                style={"height": "84vh"},
                figure={"data": figure.traces, "layout": figure.layout},
                id=get_uuid(LayoutElements.MAP_GRAPH),
            )
        ]

    @app.callback(
        Output(get_uuid(LayoutElements.FORMATIONS_GRAPH), "children"),
        Input(get_uuid(LayoutElements.FORMATIONS_WELL), "value"),
        Input(get_uuid(LayoutElements.FORMATIONS_DATE), "value"),
        Input(get_uuid(LayoutElements.FORMATIONS_ENSEMBLE), "value"),
        Input(get_uuid(LayoutElements.FORMATIONS_LINETYPE), "value"),
        Input(get_uuid(LayoutElements.FORMATIONS_DEPTHOPTION), "value"),
    )
    def _update_formation_plot(
        well: str, date: str, ensembles: List[str], linetype: str, depth_option: str
    ) -> Union[str, List[wcc.Graph]]:

        if not ensembles:
            return "No ensembles selected"

        if date not in datamodel.date_in_well(well):
            raise PreventUpdate

        figure = FormationFigure(
            well=well,
            ertdf=datamodel.ertdatadf,
            enscolors=datamodel.enscolors,
            depth_option=depth_option,
            date=date,
            ensembles=ensembles,
            simdf=datamodel.simdf,
            obsdf=datamodel.obsdatadf,
        )

        if datamodel.formations is not None:
            figure.add_formation(datamodel.formationdf)

        figure.add_simulated_lines(linetype)
        figure.add_additional_observations()
        figure.add_ert_observed()

        return [
            wcc.Graph(
                style={"height": "84vh"},
                figure=figure.figure,
            )
        ]

    @app.callback(
        Output(get_uuid(LayoutElements.FORMATIONS_LINETYPE), "options"),
        Output(get_uuid(LayoutElements.FORMATIONS_LINETYPE), "value"),
        Input(get_uuid(LayoutElements.FORMATIONS_DEPTHOPTION), "value"),
        State(get_uuid(LayoutElements.FORMATIONS_LINETYPE), "value"),
        State(get_uuid(LayoutElements.FORMATIONS_WELL), "value"),
        State(get_uuid(LayoutElements.FORMATIONS_DATE), "value"),
    )
    def _update_linetype(
        depth_option: str,
        current_linetype: str,
        current_well: str,
        current_date: str,
    ) -> Tuple[List[Dict[str, str]], str]:
        if datamodel.simdf is not None:
            df = filter_frame(
                datamodel.simdf,
                {"WELL": current_well, "DATE": current_date},
            )
            if depth_option == "TVD" or (
                depth_option == "MD"
                and "CONMD" in datamodel.simdf
                and len(df["CONMD"].unique()) == len(df["DEPTH"].unique())
            ):

                return [
                    {
                        "label": "Realization lines",
                        "value": "realization",
                    },
                    {
                        "label": "Statistical fanchart",
                        "value": "fanchart",
                    },
                ], current_linetype

        return [
            {
                "label": "Realization lines",
                "value": "realization",
            },
        ], "realization"

    @app.callback(
        Output(get_uuid(LayoutElements.FORMATIONS_DATE), "options"),
        Output(get_uuid(LayoutElements.FORMATIONS_DATE), "value"),
        Input(get_uuid(LayoutElements.FORMATIONS_WELL), "value"),
        State(get_uuid(LayoutElements.FORMATIONS_DATE), "value"),
    )
    def _update_date(well: str, current_date: str) -> Tuple[List[Dict[str, str]], str]:
        dates = datamodel.date_in_well(well)
        available_dates = [{"label": date, "value": date} for date in dates]
        date = current_date if current_date in dates else dates[0]
        return available_dates, date

    @app.callback(
        Output(get_uuid(LayoutElements.MISFITPLOT_GRAPH), "children"),
        Input(get_uuid(LayoutElements.FILTER_WELLS["misfitplot"]), "value"),
        Input(get_uuid(LayoutElements.FILTER_ZONES["misfitplot"]), "value"),
        Input(get_uuid(LayoutElements.FILTER_DATES["misfitplot"]), "value"),
        Input(get_uuid(LayoutElements.FILTER_ENSEMBLES["misfitplot"]), "value"),
    )
    def _misfit_plot(
        wells: List[str], zones: List[str], dates: List[str], ensembles: List[str]
    ) -> Union[str, List[wcc.Graph]]:
        df = filter_frame(
            datamodel.ertdatadf,
            {"WELL": wells, "ZONE": zones, "DATE": dates, "ENSEMBLE": ensembles},
        )
        if df.empty:
            return "No data matching the given filter criterias"

        return update_misfit_plot(df, datamodel.enscolors)

    @app.callback(
        Output(get_uuid(LayoutElements.CROSSPLOT_GRAPH), "children"),
        Input(get_uuid(LayoutElements.FILTER_WELLS["crossplot"]), "value"),
        Input(get_uuid(LayoutElements.FILTER_ZONES["crossplot"]), "value"),
        Input(get_uuid(LayoutElements.FILTER_DATES["crossplot"]), "value"),
        Input(get_uuid(LayoutElements.FILTER_ENSEMBLES["crossplot"]), "value"),
        Input(get_uuid(LayoutElements.CROSSPLOT_SIZE_BY), "value"),
        Input(get_uuid(LayoutElements.CROSSPLOT_COLOR_BY), "value"),
    )
    def _crossplot(
        wells: List[str],
        zones: List[str],
        dates: List[str],
        ensembles: List[str],
        sizeby: str,
        colorby: str,
    ) -> Union[str, List[wcc.Graph]]:
        df = filter_frame(
            datamodel.ertdatadf,
            {"WELL": wells, "ZONE": zones, "DATE": dates, "ENSEMBLE": ensembles},
        )
        if df.empty:
            return "No data matching the given filter criterias"
        return update_crossplot(df, sizeby, colorby)

    @app.callback(
        Output(get_uuid(LayoutElements.ERRORPLOT_GRAPH), "children"),
        Input(get_uuid(LayoutElements.FILTER_WELLS["errorplot"]), "value"),
        Input(get_uuid(LayoutElements.FILTER_ZONES["errorplot"]), "value"),
        Input(get_uuid(LayoutElements.FILTER_DATES["errorplot"]), "value"),
        Input(get_uuid(LayoutElements.FILTER_ENSEMBLES["errorplot"]), "value"),
    )
    def _errorplot(
        wells: List[str], zones: List[str], dates: List[str], ensembles: List[str]
    ) -> Union[str, List[wcc.Graph]]:
        df = filter_frame(
            datamodel.ertdatadf,
            {"WELL": wells, "ZONE": zones, "DATE": dates, "ENSEMBLE": ensembles},
        )
        if df.empty:
            return "No data matching the given filter criterias"
        return [update_errorplot(df, datamodel.enscolors)]



from typing import Callable, List

import webviz_core_components as wcc
from dash import html

from ._business_logic import RftPlotterDataModel


# pylint: disable = too-few-public-methods
class LayoutElements:
    MAP = "map-wrapper"
    MAP_GRAPH = "map-graph"
    FORMATIONS_GRAPH = "formations-graph-wrapper"
    MISFITPLOT_GRAPH = "misfit-graph-wrapper"
    CROSSPLOT_GRAPH = "crossplot-graph-wrapper"
    ERRORPLOT_GRAPH = "errorplot-graph-wrapper"
    FORMATIONS_ENSEMBLE = "formations-ensemble"
    FORMATIONS_WELL = "formations-well"
    FORMATIONS_DATE = "formations-date"
    FORMATIONS_LINETYPE = "formations-linetype"
    FORMATIONS_DEPTHOPTION = "formations-depthoption"
    MAP_ENSEMBLE = "map-ensemble"
    MAP_SIZE_BY = "map-size-by"
    MAP_COLOR_BY = "map-color-by"
    MAP_DATE_RANGE = "map-date-range"
    FILTER_ENSEMBLES = {
        "misfitplot": "ensembles-misfit",
        "crossplot": "ensembles-crossplot",
        "errorplot": "ensembles-errorplot",
    }
    FILTER_WELLS = {
        "misfitplot": "well-misfit",
        "crossplot": "well-crossplot",
        "errorplot": "well-errorplot",
    }
    FILTER_ZONES = {
        "misfitplot": "zones-misfit",
        "crossplot": "zones-crossplot",
        "errorplot": "zones-errorplot",
    }
    FILTER_DATES = {
        "misfitplot": "dates-misfit",
        "crossplot": "dates-crossplot",
        "errorplot": "dates-errorplot",
    }
    CROSSPLOT_COLOR_BY = "crossplot-color-by"
    CROSSPLOT_SIZE_BY = "crossplot-size-by"
    PARAMRESP_ENSEMBLE = "param-response-ensemble"
    PARAMRESP_WELL = "param-response-well"
    PARAMRESP_DATE = "param-response-date"
    PARAMRESP_ZONE = "param-response-zone"
    PARAMRESP_PARAM = "param-response-param"
    PARAMRESP_CORRTYPE = "param-response-corrtype"
    PARAMRESP_CORR_BARCHART = "paramresp-corr-barchart"
    PARAMRESP_CORR_BARCHART_FIGURE = "paramresp-corr-barchart-figure"
    PARAMRESP_SCATTERPLOT = "paramresp-scatterplot"
    PARAMRESP_FORMATIONS = "paramresp-formations"
    PARAMRESP_DATE_DROPDOWN = "paramresp-well-dropdown"
    PARAMRESP_ZONE_DROPDOWN = "paramresp-zone-dropdown"


def main_layout(get_uuid: Callable, datamodel: RftPlotterDataModel) -> wcc.Tabs:

    tabs = [
        wcc.Tab(
            label="RFT Map",
            children=[
                wcc.FlexBox(
                    children=[
                        wcc.Frame(
                            style={"flex": 1, "height": "87vh"},
                            children=[
                                map_plot_selectors(get_uuid, datamodel),
                                formation_plot_selectors(get_uuid, datamodel),
                            ],
                        ),
                        wcc.Frame(
                            style={"flex": 3, "height": "87vh"},
                            color="white",
                            highlight=False,
                            id=get_uuid(LayoutElements.MAP),
                            children=[],
                        ),
                        wcc.Frame(
                            style={"flex": 3, "height": "87vh"},
                            color="white",
                            highlight=False,
                            id=get_uuid(LayoutElements.FORMATIONS_GRAPH),
                            children=[],
                        ),
                    ]
                )
            ],
        ),
        wcc.Tab(
            label="RFT misfit per real",
            children=[
                wcc.FlexBox(
                    children=[
                        wcc.Frame(
                            style={"flex": 1, "height": "87vh"},
                            children=filter_layout(get_uuid, datamodel, "misfitplot"),
                        ),
                        wcc.Frame(
                            style={"flex": 6, "height": "87vh"},
                            color="white",
                            highlight=False,
                            id=get_uuid(LayoutElements.MISFITPLOT_GRAPH),
                            children=[],
                        ),
                    ]
                )
            ],
        ),
        wcc.Tab(
            label="RFT crossplot - sim vs obs",
            children=[
                wcc.FlexBox(
                    children=[
                        wcc.Frame(
                            style={"flex": 1, "height": "87vh"},
                            children=[
                                filter_layout(get_uuid, datamodel, "crossplot"),
                                size_color_layout(get_uuid),
                            ],
                        ),
                        wcc.Frame(
                            style={"flex": 6, "height": "87vh"},
                            color="white",
                            highlight=False,
                            children=[
                                html.Div(id=get_uuid(LayoutElements.CROSSPLOT_GRAPH)),
                            ],
                        ),
                    ],
                ),
            ],
        ),
        wcc.Tab(
            label="RFT misfit per observation",
            children=[
                wcc.FlexBox(
                    children=[
                        wcc.Frame(
                            style={"flex": 1, "height": "87vh"},
                            children=filter_layout(get_uuid, datamodel, "errorplot"),
                        ),
                        wcc.Frame(
                            color="white",
                            highlight=False,
                            style={"flex": 6, "height": "87vh"},
                            id=get_uuid(LayoutElements.ERRORPLOT_GRAPH),
                            children=[],
                        ),
                    ],
                ),
            ],
        ),
    ]

    # It this is not a sensitivity run, add the parameter response tab
    if not datamodel.param_model.sensrun:
        tabs.append(
            wcc.Tab(
                label="RFT parameter response",
                children=parameter_response_layout(
                    get_uuid=get_uuid, datamodel=datamodel
                ),
            )
        )

    return wcc.Tabs(children=tabs)


def parameter_response_selector_layout(
    get_uuid: Callable, datamodel: RftPlotterDataModel
) -> wcc.Frame:
    ensembles = datamodel.ensembles
    well_names = datamodel.well_names
    params = datamodel.parameters if not datamodel.parameters is None else []
    return wcc.Frame(
        style={
            "height": "87vh",
            "overflowY": "auto",
            "font-size": "15px",
        },
        children=[
            wcc.Selectors(
                label="Selections",
                children=[
                    wcc.Dropdown(
                        label="Ensemble",
                        id=get_uuid(LayoutElements.PARAMRESP_ENSEMBLE),
                        options=[{"label": ens, "value": ens} for ens in ensembles],
                        value=ensembles[0],
                        clearable=False,
                    ),
                    wcc.Dropdown(
                        label="Well",
                        id=get_uuid(LayoutElements.PARAMRESP_WELL),
                        options=[{"label": well, "value": well} for well in well_names],
                        value=well_names[0] if well_names else "",
                        clearable=False,
                    ),
                    html.Div(
                        id=get_uuid(LayoutElements.PARAMRESP_DATE_DROPDOWN),
                        children=wcc.Dropdown(
                            label="Date",
                            id=get_uuid(LayoutElements.PARAMRESP_DATE),
                            options=None,
                            value=None,
                            clearable=False,
                        ),
                    ),
                    html.Div(
                        id=get_uuid(LayoutElements.PARAMRESP_ZONE_DROPDOWN),
                        children=wcc.Dropdown(
                            label="Zone",
                            id=get_uuid(LayoutElements.PARAMRESP_ZONE),
                            options=None,
                            clearable=False,
                            value=None,
                        ),
                    ),
                    wcc.Dropdown(
                        label="Parameter",
                        id=get_uuid(LayoutElements.PARAMRESP_PARAM),
                        options=[{"label": param, "value": param} for param in params],
                        clearable=False,
                        value=None,
                    ),
                ],
            ),
            wcc.Selectors(
                label="Options",
                children=[
                    wcc.RadioItems(
                        label="Correlation options",
                        id=get_uuid(LayoutElements.PARAMRESP_CORRTYPE),
                        options=[
                            {
                                "label": "Simulated vs parameters",
                                "value": "sim_vs_param",
                            },
                            {
                                "label": "Parameter vs simulated",
                                "value": "param_vs_sim",
                            },
                        ],
                        value="sim_vs_param",
                    ),
                ],
            ),
        ],
    )


def parameter_response_layout(
    get_uuid: Callable, datamodel: RftPlotterDataModel
) -> wcc.FlexBox:
    return wcc.FlexBox(
        children=[
            wcc.FlexColumn(
                flex=1, children=parameter_response_selector_layout(get_uuid, datamodel)
            ),
            wcc.FlexColumn(
                flex=4,
                children=wcc.FlexBox(
                    children=[
                        wcc.FlexColumn(
                            flex=2,
                            children=[
                                wcc.Frame(
                                    style={"height": "41.5vh"},
                                    id=get_uuid(LayoutElements.PARAMRESP_CORR_BARCHART),
                                    color="white",
                                    highlight=False,
                                    children=[],
                                ),
                                wcc.Frame(
                                    style={"height": "41.5vh"},
                                    id=get_uuid(LayoutElements.PARAMRESP_SCATTERPLOT),
                                    color="white",
                                    highlight=False,
                                    children=[],
                                ),
                            ],
                        ),
                        wcc.FlexColumn(
                            flex=2,
                            children=[
                                wcc.Frame(
                                    id=get_uuid(LayoutElements.PARAMRESP_FORMATIONS),
                                    color="white",
                                    highlight=False,
                                    style={"height": "87vh"},
                                    children=[],
                                )
                            ],
                        ),
                    ],
                ),
            ),
        ]
    )


def formation_plot_selectors(
    get_uuid: Callable, datamodel: RftPlotterDataModel
) -> List[html.Div]:
    ensembles = datamodel.ensembles
    well_names = datamodel.well_names
    date_in_well = datamodel.date_in_well
    return wcc.Selectors(
        label="Formation plot settings",
        children=[
            wcc.Dropdown(
                label="Ensemble",
                id=get_uuid(LayoutElements.FORMATIONS_ENSEMBLE),
                options=[{"label": ens, "value": ens} for ens in ensembles],
                value=ensembles[0],
                multi=True,
                clearable=False,
            ),
            wcc.Dropdown(
                label="Well",
                id=get_uuid(LayoutElements.FORMATIONS_WELL),
                options=[{"label": well, "value": well} for well in well_names],
                value=well_names[0],
                clearable=False,
            ),
            wcc.Dropdown(
                label="Date",
                id=get_uuid(LayoutElements.FORMATIONS_DATE),
                options=[
                    {"label": date, "value": date}
                    for date in date_in_well(well_names[0])
                ],
                clearable=False,
                value=date_in_well(well_names[0])[0],
            ),
            wcc.RadioItems(
                label="Plot simulations as",
                id=get_uuid(LayoutElements.FORMATIONS_LINETYPE),
                options=[
                    {
                        "label": "Realization lines",
                        "value": "realization",
                    },
                    {
                        "label": "Statistical fanchart",
                        "value": "fanchart",
                    },
                ],
                value="realization",
            ),
            wcc.RadioItems(
                label="Depth option",
                id=get_uuid(LayoutElements.FORMATIONS_DEPTHOPTION),
                options=[
                    {
                        "label": "TVD",
                        "value": "TVD",
                    },
                    {
                        "label": "MD",
                        "value": "MD",
                    },
                ],
                value="TVD",
            ),
        ],
    )


def map_plot_selectors(
    get_uuid: Callable, datamodel: RftPlotterDataModel
) -> List[html.Div]:
    ensembles = datamodel.ensembles
    return wcc.Selectors(
        label="Map plot settings",
        children=[
            wcc.Dropdown(
                label="Ensemble",
                id=get_uuid(LayoutElements.MAP_ENSEMBLE),
                options=[{"label": ens, "value": ens} for ens in ensembles],
                value=ensembles[0],
                clearable=False,
            ),
            wcc.Dropdown(
                label="Size points by",
                id=get_uuid(LayoutElements.MAP_SIZE_BY),
                options=[
                    {
                        "label": "Standard Deviation",
                        "value": "STDDEV",
                    },
                    {
                        "label": "Misfit",
                        "value": "ABSDIFF",
                    },
                ],
                value="ABSDIFF",
                clearable=False,
            ),
            wcc.Dropdown(
                label="Color points by",
                id=get_uuid(LayoutElements.MAP_COLOR_BY),
                options=[
                    {
                        "label": "Misfit",
                        "value": "ABSDIFF",
                    },
                    {
                        "label": "Standard Deviation",
                        "value": "STDDEV",
                    },
                    {
                        "label": "Year",
                        "value": "YEAR",
                    },
                ],
                value="STDDEV",
                clearable=False,
            ),
            wcc.RangeSlider(
                label="Filter date range",
                id=get_uuid(LayoutElements.MAP_DATE_RANGE),
                min=datamodel.ertdatadf["DATE_IDX"].min(),
                max=datamodel.ertdatadf["DATE_IDX"].max(),
                value=[
                    datamodel.ertdatadf["DATE_IDX"].min(),
                    datamodel.ertdatadf["DATE_IDX"].max(),
                ],
                marks=datamodel.date_marks,
            ),
        ],
    )


def filter_layout(
    get_uuid: Callable, datamodel: RftPlotterDataModel, tab: str
) -> List[wcc.Selectors]:
    """Layout for shared filters"""
    ensembles = datamodel.ensembles
    well_names = datamodel.well_names
    zone_names = datamodel.zone_names
    dates = datamodel.dates
    return wcc.Selectors(
        label="Selectors",
        children=[
            wcc.SelectWithLabel(
                label="Ensembles",
                size=min(4, len(ensembles)),
                id=get_uuid(LayoutElements.FILTER_ENSEMBLES[tab]),
                options=[{"label": name, "value": name} for name in ensembles],
                value=ensembles,
                multi=True,
            ),
            wcc.SelectWithLabel(
                label="Wells",
                size=min(20, len(well_names)),
                id=get_uuid(LayoutElements.FILTER_WELLS[tab]),
                options=[{"label": name, "value": name} for name in well_names],
                value=well_names,
                multi=True,
            ),
            wcc.SelectWithLabel(
                label="Zones",
                size=min(10, len(zone_names)),
                id=get_uuid(LayoutElements.FILTER_ZONES[tab]),
                options=[{"label": name, "value": name} for name in zone_names],
                value=zone_names,
                multi=True,
            ),
            wcc.SelectWithLabel(
                label="Dates",
                size=min(10, len(dates)),
                id=get_uuid(LayoutElements.FILTER_DATES[tab]),
                options=[{"label": name, "value": name} for name in dates],
                value=dates,
                multi=True,
            ),
        ],
    )


def size_color_layout(get_uuid: Callable) -> List[html.Div]:
    return wcc.Selectors(
        label="Plot settings",
        children=[
            wcc.Dropdown(
                label="Color by",
                id=get_uuid(LayoutElements.CROSSPLOT_COLOR_BY),
                options=[
                    {
                        "label": "Misfit",
                        "value": "ABSDIFF",
                    },
                    {
                        "label": "Standard Deviation",
                        "value": "STDDEV",
                    },
                ],
                value="STDDEV",
                clearable=False,
            ),
            wcc.Dropdown(
                label="Size by",
                id=get_uuid(LayoutElements.CROSSPLOT_SIZE_BY),
                options=[
                    {
                        "label": "Standard Deviation",
                        "value": "STDDEV",
                    },
                    {
                        "label": "Misfit",
                        "value": "ABSDIFF",
                    },
                ],
                value="ABSDIFF",
                clearable=False,
            ),
        ],
    )



import datetime
from typing import Optional

import numpy as np
import pandas as pd

from webviz_subsurface._providers import Frequency
from webviz_subsurface._utils.dataframe_utils import (
    assert_date_column_is_datetime_object,
    make_date_column_datetime_object,
)

###################################################################################
# NOTE: This code is a copy and modification of
# webviz_subsurface/_datainput/from_timeseries_cumulatives.py for usage with
# EnsembleSummaryProvider functionality
#
# Renaming is performed for clarity
#
# Some additional functions are added.
###################################################################################


def is_interval_or_average_vector(vector: str) -> bool:
    return vector.startswith("AVG_") or vector.startswith("INTVL_")


def get_cumulative_vector_name(vector: str) -> str:
    if not is_interval_or_average_vector(vector):
        raise ValueError(
            f'Expected "{vector}" to be a vector calculated from cumulative!'
        )

    if vector.startswith("AVG_"):
        return f"{vector[4:7] + vector[7:].replace('R', 'T', 1)}"
    if vector.startswith("INTVL_"):
        return vector.lstrip("INTVL_")
    raise ValueError(f"Expected {vector} to be a cumulative vector!")


def calculate_from_resampled_cumulative_vectors_df(
    vectors_df: pd.DataFrame,
    as_rate_per_day: bool,
) -> pd.DataFrame:
    """
    Calculates interval delta or average rate data for vector columns in provided dataframe.
    This function assumes data is already resampled when retrieved with ensemble summary
    provider.

    Date sampling should be according to Frequency enum in
    webviz_subsurface/_providers/ensemble_summary_provider.py.
    Does not handle raw data and varying time delta less than daily!

    `INPUT:`
    * vectors_df: pd.Dataframe - Dataframe with columns:
        ["DATE", "REAL", vector1, ..., vectorN]

    `NOTE:`
    - Does not handle raw data format with varying sampling or sampling frequency higher than
    daily!
    - Dataframe has columns:\n
    "DATE": Series with dates on datetime.datetime format
    "REAL": Series of realization number identifier
    vector1, ..., vectorN: Series of vector data for vector of given column name

    `TODO:`
    * IMPROVE FUNCTION NAME?
    * Handle raw data format?
    * Give e.g. a dict with info of "avg and intvl" calculation for each vector column?
    Can thereby calculate everything for provided vector columns and no iterate column per
    column?
    """
    assert_date_column_is_datetime_object(vectors_df)

    vectors_df = vectors_df.copy()

    column_keys = [elm for elm in vectors_df.columns if elm not in ["DATE", "REAL"]]

    # Sort by realizations, thereafter dates
    vectors_df.sort_values(by=["REAL", "DATE"], inplace=True)

    # Create column of unique id for realizations. .diff() takes diff between an index
    # and previous index in a column. Thereby if "realuid" is != 0, the .diff() is
    # between two realizations.
    # Could alternatively loop over ensembles and realizations, but this is quicker for
    # larger datasets.
    vectors_df["realuid"] = vectors_df["REAL"]

    vectors_df.set_index(["DATE"], inplace=True)

    cumulative_name_map = {
        vector: rename_vector_from_cumulative(vector, as_rate_per_day)
        for vector in column_keys
    }
    cumulative_vectors = list(cumulative_name_map.values())

    # Take diff of given column_keys indexes - preserve REAL
    cumulative_vectors_df = pd.concat(
        [
            vectors_df[["REAL"]],
            vectors_df[["realuid"] + column_keys]
            .diff()
            .shift(-1)
            .rename(
                mapper=cumulative_name_map,
                axis=1,
            ),
        ],
        axis=1,
    )
    cumulative_vectors_df[cumulative_vectors] = cumulative_vectors_df[
        cumulative_vectors
    ].fillna(value=0)

    # Reset index (DATE becomes regular column)
    cumulative_vectors_df.reset_index(inplace=True)

    # Convert interval cumulative to daily average rate if requested
    if as_rate_per_day:
        days = cumulative_vectors_df["DATE"].diff().shift(-1).dt.days.fillna(value=0)
        for vector in column_keys:
            with np.errstate(invalid="ignore"):
                cumulative_vector_name = cumulative_name_map[vector]
                cumulative_vectors_df.loc[:, cumulative_vector_name] = (
                    cumulative_vectors_df[cumulative_vector_name].values / days.values
                )

    # Find .diff() between two realizations and set value = 0
    cumulative_vectors_df.loc[
        cumulative_vectors_df["realuid"] != 0, cumulative_vectors
    ] = 0
    cumulative_vectors_df.drop("realuid", axis=1, inplace=True)

    make_date_column_datetime_object(cumulative_vectors_df)

    return cumulative_vectors_df


def rename_vector_from_cumulative(vector: str, as_rate: bool) -> str:
    """This function assumes that it is a cumulative/total vector named in the Eclipse standard
    and is fairly naive when converting to rate. Based in the list in libecl
    https://github.com/equinor/libecl/blob/69f1ee0ddf696c87b6d85eca37eed7e8b66ac2db/\
        lib/ecl/smspec_node.cpp#L531-L586
    the T identifying total/cumulative should not occur before letter 4,
    as all the listed strings are prefixed with one or two letters in the vectors.
    Therefore starting the replace at the position 3 (4th letter) to reduce risk of errors
    in the conversion to rate naming, but hard to be completely safe.
    """
    return (
        f"AVG_{vector[0:3] + vector[3:].replace('T', 'R', 1)}"
        if as_rate
        else f"INTVL_{vector}"
    )


# pylint: disable=too-many-return-statements
def datetime_to_intervalstr(date: datetime.datetime, freq: Frequency) -> Optional[str]:
    if date is None:
        return None

    if freq == Frequency.DAILY:
        return f"{date.year}-{date.month:02d}-{date.day:02d}"
    if freq == Frequency.WEEKLY:
        # Note: weekyear may differ from actual year for week numbers 1,52,53.
        (weekyear, week, _) = date.isocalendar()
        return f"{weekyear}-W{week:02d}"
    if freq == Frequency.MONTHLY:
        return f"{date.year}-{date.month:02d}"
    if freq == Frequency.QUARTERLY:
        return f"{date.year}-Q{1 + (date.month-1)//3}"
    if freq == Frequency.YEARLY:
        return f"{date.year}"

    # Using isoformat if frequency is none of the listed instead of error.
    return date.isoformat()



from .derived_delta_ensemble_vectors_accessor_impl import (
    DerivedDeltaEnsembleVectorsAccessorImpl,
)
from .derived_ensemble_vectors_accessor_impl import DerivedEnsembleVectorsAccessorImpl
from .derived_vectors_accessor import DerivedVectorsAccessor
from .provider_set import ProviderSet
from .types import (
    DeltaEnsemble,
    FanchartOptions,
    StatisticsFromOptions,
    StatisticsOptions,
    SubplotGroupByOptions,
    TraceOptions,
    VisualizationOptions,
)



import copy
import warnings
from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple

import dash
import webviz_core_components as wcc
from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.webviz_assets import WEBVIZ_ASSETS

import webviz_subsurface
from webviz_subsurface._abbreviations.reservoir_simulation import (
    historical_vector,
    simulation_vector_description,
)
from webviz_subsurface._providers import Frequency
from webviz_subsurface._utils.simulation_timeseries import (
    check_and_format_observations,
    set_simulation_line_shape_fallback,
)
from webviz_subsurface._utils.vector_calculator import (
    add_expressions_to_vector_selector_data,
    expressions_from_config,
    validate_predefined_expression,
)
from webviz_subsurface._utils.vector_selector import add_vector_to_vector_selector_data
from webviz_subsurface._utils.webvizstore_functions import get_path

from ._callbacks import plugin_callbacks
from ._layout import LayoutElements, main_layout
from .types import VisualizationOptions
from .types.provider_set import (
    create_lazy_provider_set_from_paths,
    create_presampled_provider_set_from_paths,
)
from .utils.from_timeseries_cumulatives import rename_vector_from_cumulative


class SimulationTimeSeries(WebvizPluginABC):
    # pylint: disable=too-many-arguments,too-many-locals,too-many-statements
    def __init__(
        self,
        app: dash.Dash,
        webviz_settings: WebvizSettings,
        ensembles: Optional[list] = None,
        rel_file_pattern: str = "share/results/unsmry/*.arrow",
        perform_presampling: bool = False,
        obsfile: Path = None,
        options: dict = None,
        sampling: str = Frequency.MONTHLY.value,
        predefined_expressions: str = None,
        line_shape_fallback: str = "linear",
    ) -> None:
        super().__init__()

        # NOTE: Temporary css, pending on new wcc modal component.
        # See: https://github.com/equinor/webviz-core-components/issues/163
        WEBVIZ_ASSETS.add(
            Path(webviz_subsurface.__file__).parent / "_assets" / "css" / "modal.css"
        )

        self._webviz_settings = webviz_settings
        self._obsfile = obsfile

        self._line_shape_fallback = set_simulation_line_shape_fallback(
            line_shape_fallback
        )

        # Must define valid freqency!
        if Frequency.from_string_value(sampling) is None:
            raise ValueError(
                'Sampling frequency conversion is "None", i.e. Raw sampling, and '
                "is not supported by plugin yet!"
            )
        self._sampling = Frequency(sampling)
        self._presampled_frequency = None

        # TODO: Update functionality when allowing raw data and csv file input
        # NOTE: If csv is implemented-> handle/disable statistics, INTVL_, AVG_, delta
        # ensemble, etc.
        if ensembles is not None:
            ensemble_paths: Dict[str, Path] = {
                ensemble_name: webviz_settings.shared_settings["scratch_ensembles"][
                    ensemble_name
                ]
                for ensemble_name in ensembles
            }
            if perform_presampling:
                self._presampled_frequency = self._sampling
                self._input_provider_set = create_presampled_provider_set_from_paths(
                    ensemble_paths, rel_file_pattern, self._presampled_frequency
                )
            else:
                self._input_provider_set = create_lazy_provider_set_from_paths(
                    ensemble_paths, rel_file_pattern
                )
        else:
            raise ValueError('Incorrect argument, must provide "ensembles"')

        if not self._input_provider_set:
            raise ValueError(
                "Initial provider set is undefined, and ensemble summary providers"
                " are not instanciated for plugin"
            )

        self._theme = webviz_settings.theme

        self._observations = {}
        if self._obsfile:
            self._observations = check_and_format_observations(get_path(self._obsfile))

        # NOTE: Initially keep set of all vector names - can make dynamic if wanted?
        vector_names = self._input_provider_set.all_vector_names()
        non_historical_vector_names = [
            vector
            for vector in vector_names
            if historical_vector(vector, None, False) not in vector_names
        ]

        # NOTE: Initially: With set of vector names, the vector selector data is static
        # Can be made dynamic based on selected ensembles - i.e. vectors present among
        # selected providers?
        self._vector_selector_base_data: list = []
        self._vector_calculator_data: list = []
        for vector in non_historical_vector_names:
            split = vector.split(":")
            add_vector_to_vector_selector_data(
                self._vector_selector_base_data,
                vector,
                simulation_vector_description(split[0]),
            )
            add_vector_to_vector_selector_data(
                self._vector_calculator_data,
                vector,
                simulation_vector_description(split[0]),
            )

            metadata = (
                self._input_provider_set.vector_metadata(vector)
                if self._input_provider_set
                else None
            )
            if metadata and metadata.is_total:
                # Get the likely name for equivalent rate vector and make dropdown options.
                # Requires that the time_index was either defined or possible to infer.
                avgrate_vec = rename_vector_from_cumulative(vector=vector, as_rate=True)
                interval_vec = rename_vector_from_cumulative(
                    vector=vector, as_rate=False
                )

                avgrate_split = avgrate_vec.split(":")
                interval_split = interval_vec.split(":")

                add_vector_to_vector_selector_data(
                    self._vector_selector_base_data,
                    avgrate_vec,
                    f"{simulation_vector_description(avgrate_split[0])} ({avgrate_vec})",
                )
                add_vector_to_vector_selector_data(
                    self._vector_selector_base_data,
                    interval_vec,
                    f"{simulation_vector_description(interval_split[0])} ({interval_vec})",
                )

        # Retreive predefined expressions from configuration and validate
        self._predefined_expressions_path = (
            None
            if predefined_expressions is None
            else webviz_settings.shared_settings["predefined_expressions"][
                predefined_expressions
            ]
        )
        self._predefined_expressions = expressions_from_config(
            get_path(self._predefined_expressions_path)
            if self._predefined_expressions_path
            else None
        )
        for expression in self._predefined_expressions:
            valid, message = validate_predefined_expression(
                expression, self._vector_selector_base_data
            )
            if not valid:
                warnings.warn(message)
            expression["isValid"] = valid

        # Create initial vector selector data with predefined expressions
        self._initial_vector_selector_data = copy.deepcopy(
            self._vector_selector_base_data
        )
        add_expressions_to_vector_selector_data(
            self._initial_vector_selector_data, self._predefined_expressions
        )

        plot_options = options if options else {}
        self._initial_visualization_selection = VisualizationOptions(
            plot_options.get("visualization", "statistics")
        )
        self._initial_vectors: List[str] = []
        if "vectors" not in plot_options:
            self._initial_vectors = []
        for vector in [
            vector
            for vector in ["vector1", "vector2", "vector3"]
            if vector in plot_options
        ]:
            self._initial_vectors.append(plot_options[vector])
        self._initial_vectors = self._initial_vectors[:3]

        # Set callbacks
        self.set_callbacks(app)

    @property
    def layout(self) -> wcc.FlexBox:
        return main_layout(
            get_uuid=self.uuid,
            ensemble_names=self._input_provider_set.names(),
            vector_selector_data=self._initial_vector_selector_data,
            vector_calculator_data=self._vector_calculator_data,
            predefined_expressions=self._predefined_expressions,
            realizations=self._input_provider_set.all_realizations(),
            disable_resampling_dropdown=self._presampled_frequency is not None,
            selected_resampling_frequency=self._sampling,
            selected_visualization=self._initial_visualization_selection,
            selected_vectors=self._initial_vectors,
        )

    def set_callbacks(self, app: dash.Dash) -> None:
        plugin_callbacks(
            app=app,
            get_uuid=self.uuid,
            get_data_output=self.plugin_data_output,
            get_data_requested=self.plugin_data_requested,
            input_provider_set=self._input_provider_set,
            theme=self._theme,
            initial_selected_vectors=self._initial_vectors,
            vector_selector_base_data=self._vector_selector_base_data,
            observations=self._observations,
            line_shape_fallback=self._line_shape_fallback,
        )

    @property
    def tour_steps(self) -> List[dict]:
        return [
            {
                "id": self.uuid(LayoutElements.TOUR_STEP_MAIN_LAYOUT),
                "content": "Dashboard displaying reservoir simulation time series.",
            },
            {
                "id": self.uuid(LayoutElements.GRAPH),
                "content": (
                    "Visualization of selected time series. "
                    "Different options can be set in the menu to the left."
                ),
            },
            {
                "id": self.uuid(LayoutElements.TOUR_STEP_SETTINGS_LAYOUT),
                "content": (
                    "Settings to configure data and layout of the time series visualization."
                ),
            },
            {
                "id": self.uuid(LayoutElements.TOUR_STEP_GROUP_BY),
                "content": (
                    "Setting to group visualization data according to selection. "
                    "Subplot per selected vector or per selected ensemble."
                ),
            },
            {
                "id": self.uuid(LayoutElements.RESAMPLING_FREQUENCY_DROPDOWN),
                "content": (
                    "Select resampling frequency for the time series data. "
                    "With presampled data, the dropdown is disabled and the presampling "
                    "frequency shown."
                ),
            },
            {
                "id": self.uuid(LayoutElements.ENSEMBLES_DROPDOWN),
                "content": (
                    "Display time series from one or several ensembles. "
                    "Ensembles will be overlain in subplot or represented per subplot, "
                    'based on selection in "Group By".'
                ),
            },
            {
                "id": self.uuid(LayoutElements.TOUR_STEP_DELTA_ENSEMBLE),
                "content": (
                    "Create delta ensembles (A-B). "
                    "Define delta between two ensembles and make available among "
                    "selectable ensembles."
                ),
            },
            {
                "id": self.uuid(LayoutElements.VECTOR_SELECTOR),
                "content": (
                    "Display up to three different time series. "
                    "Each time series will be visualized in a separate plot. "
                    "Vectors prefixed with AVG_ and INTVL_ are calculated in the fly "
                    "from cumulative vectors, providing average rates and interval cumulatives "
                    "over a time interval from the selected resampling frequency. Vectors "
                    "categorized as calculated are created using the Vector Calculator below."
                ),
            },
            {
                "id": self.uuid(LayoutElements.VECTOR_CALCULATOR_OPEN_BUTTON),
                "content": (
                    "Create mathematical expressions with provided vector time series. "
                    "Parsing of the mathematical expression is handled and will give feedback "
                    "when entering invalid expressions. "
                    "The expressions are calculated on the fly and can be selected among the time "
                    "series to be shown in the visualization."
                ),
            },
            {
                "id": self.uuid(LayoutElements.VISUALIZATION_RADIO_ITEMS),
                "content": (
                    "Choose between different visualizations. 1. Show time series as "
                    "individual lines per realization. 2. Show statistical lines per "
                    "ensemble. 3. Show statistical fanchart per ensemble."
                ),
            },
            {
                "id": self.uuid(LayoutElements.TOUR_STEP_OPTIONS),
                "content": (
                    "Various plot options: Whether to include history trace or vector observations "
                    "and which statistics to show if statistical lines or fanchart is chosen as "
                    "visualization."
                ),
            },
            {
                "id": self.uuid(LayoutElements.REALIZATIONS_FILTER_SELECTOR),
                "content": (
                    "Filter realizations. Select realization numbers to include in visualization, "
                    "and utilize in statistics calculation when calculating from selected subset."
                ),
            },
            {
                "id": self.uuid(LayoutElements.STATISTICS_FROM_RADIO_ITEMS),
                "content": (
                    "Select whether to calculate statistics from all realizations, or to calculate "
                    "statistics from the selected subset of realizations "
                ),
            },
        ]

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        functions: List[Tuple[Callable, list]] = []
        if self._obsfile:
            functions.append((get_path, [{"path": self._obsfile}]))
        if self._predefined_expressions_path:
            functions.append((get_path, [{"path": self._predefined_expressions_path}]))
        return functions



from .ensemble_summary_provider.ensemble_summary_provider import (
    EnsembleSummaryProvider,
    Frequency,
    VectorMetadata,
)
from .ensemble_summary_provider.ensemble_summary_provider_factory import (
    EnsembleSummaryProviderFactory,
)
from .ensemble_table_provider import EnsembleTableProvider, EnsembleTableProviderSet
from .ensemble_table_provider_factory import EnsembleTableProviderFactory



from typing import Any, Callable, Dict, List, Optional, Tuple

import dash
import webviz_subsurface_components
from dash.dependencies import Input, Output, State

from ._ensemble_group_tree_data import EnsembleGroupTreeData


def plugin_callbacks(
    app: dash.Dash,
    get_uuid: Callable,
    group_tree_data: Dict[str, EnsembleGroupTreeData],
) -> None:
    @app.callback(
        Output({"id": get_uuid("controls"), "element": "tree_mode"}, "options"),
        Output({"id": get_uuid("controls"), "element": "tree_mode"}, "value"),
        Output({"id": get_uuid("options"), "element": "statistical_option"}, "value"),
        Output({"id": get_uuid("options"), "element": "realization"}, "options"),
        Output({"id": get_uuid("options"), "element": "realization"}, "value"),
        Input({"id": get_uuid("controls"), "element": "ensemble"}, "value"),
        State({"id": get_uuid("controls"), "element": "tree_mode"}, "value"),
        State({"id": get_uuid("options"), "element": "statistical_option"}, "value"),
        State({"id": get_uuid("options"), "element": "realization"}, "value"),
    )
    def _update_ensemble_options(
        ensemble_name: str,
        tree_mode_state: str,
        stat_option_state: str,
        real_state: int,
    ) -> Tuple[List[Dict[str, Any]], str, str, List[Dict[str, Any]], Optional[int]]:
        """Updates the selection options when the ensemble value changes"""
        tree_mode_options: List[Dict[str, Any]] = [
            {
                "label": "Statistics",
                "value": "statistics",
            },
            {
                "label": "Single realization",
                "value": "single_real",
            },
        ]
        tree_mode_value = (
            tree_mode_state if tree_mode_state is not None else "statistics"
        )
        stat_option_value = (
            stat_option_state if stat_option_state is not None else "mean"
        )
        ensemble = group_tree_data[ensemble_name]
        if not ensemble.tree_is_equivalent_in_all_real():
            tree_mode_options[0]["label"] = "Ensemble mean (disabled)"
            tree_mode_options[0]["disabled"] = True
            tree_mode_value = "single_real"

        unique_real = ensemble.get_unique_real()

        return (
            tree_mode_options,
            tree_mode_value,
            stat_option_value,
            [{"label": real, "value": real} for real in unique_real],
            real_state if real_state in unique_real else min(unique_real),
        )

    @app.callback(
        Output(get_uuid("grouptree_wrapper"), "children"),
        Input({"id": get_uuid("controls"), "element": "tree_mode"}, "value"),
        Input({"id": get_uuid("options"), "element": "statistical_option"}, "value"),
        Input({"id": get_uuid("options"), "element": "realization"}, "value"),
        Input({"id": get_uuid("filters"), "element": "prod_inj_other"}, "value"),
        State({"id": get_uuid("controls"), "element": "ensemble"}, "value"),
    )
    def _render_grouptree(
        tree_mode: str,
        stat_option: str,
        real: int,
        prod_inj_other: list,
        ensemble_name: str,
    ) -> list:
        """This callback updates the input dataset to the Grouptree component."""
        data, edge_options, node_options = group_tree_data[
            ensemble_name
        ].create_grouptree_dataset(tree_mode, stat_option, real, prod_inj_other)

        return [
            webviz_subsurface_components.GroupTree(
                id="grouptree",
                data=data,
                edge_options=edge_options,
                node_options=node_options,
            ),
        ]

    @app.callback(
        Output(
            {"id": get_uuid("options"), "element": "statistical_options"},
            "style",
        ),
        Output(
            {"id": get_uuid("options"), "element": "single_real_options"},
            "style",
        ),
        Input(
            {"id": get_uuid("controls"), "element": "tree_mode"},
            "value",
        ),
    )
    def _show_hide_single_real_options(
        tree_mode: str,
    ) -> Tuple[Dict[str, str], Dict[str, str]]:
        if tree_mode == "statistics":
            return {"display": "block"}, {"display": "none"}
        return {"display": "none"}, {"display": "block"}



from typing import Callable, List, Optional

import dash_bootstrap_components as dbc
import webviz_core_components as wcc
import webviz_subsurface_components as wsc
from dash import dash_table, dcc, html
from webviz_subsurface_components import ExpressionInfo

from webviz_subsurface._providers import Frequency
from webviz_subsurface._utils.vector_calculator import (
    get_custom_vector_definitions_from_expressions,
)

from .types import (
    FanchartOptions,
    StatisticsFromOptions,
    StatisticsOptions,
    SubplotGroupByOptions,
    TraceOptions,
    VisualizationOptions,
)


# pylint: disable=too-few-public-methods
class LayoutElements:
    """
    Definition of names of HTML-elements in layout

    TODO: Consider ids as in AiO convention https://dash.plotly.com/all-in-one-components
    """

    GRAPH = "graph"
    GRAPH_DATA_HAS_CHANGED_TRIGGER = (
        "graph_data_has_changed_trigger"  # NOTE: To force re-render of graph
    )

    ENSEMBLES_DROPDOWN = "ensembles_dropdown"
    VECTOR_SELECTOR = "vector_selector"

    VECTOR_CALCULATOR = "vector_calculator"
    VECTOR_CALCULATOR_MODAL = "vector_calculator_modal"
    VECTOR_CALCULATOR_OPEN_BUTTON = "vector_calculator_open_button"
    VECTOR_CALCULATOR_EXPRESSIONS = "vector_calculator_expressions"
    VECTOR_CALCULATOR_EXPRESSIONS_OPEN_MODAL = (
        "vector_calculator_expressions_open_modal"
    )

    DELTA_ENSEMBLE_A_DROPDOWN = "delta_ensemble_A_dropdown"
    DELTA_ENSEMBLE_B_DROPDOWN = "delta_ensemble_B_dropdown"
    DELTA_ENSEMBLE_CREATE_BUTTON = "delta_ensemble_create_button"
    CREATED_DELTA_ENSEMBLES = "created_delta_ensemble_names"
    CREATED_DELTA_ENSEMBLE_NAMES_TABLE = "created_delta_ensemble_names_table"
    CREATED_DELTA_ENSEMBLE_NAMES_TABLE_COLUMN = (
        "created_delta_ensemble_names_table_column"
    )

    VISUALIZATION_RADIO_ITEMS = "visualization_radio_items"

    PLOT_FANCHART_OPTIONS_CHECKLIST = "plot_fanchart_options_checklist"
    PLOT_STATISTICS_OPTIONS_CHECKLIST = "plot_statistics_options_checklist"
    PLOT_TRACE_OPTIONS_CHECKLIST = "plot_trace_options_checklist"

    SUBPLOT_OWNER_OPTIONS_RADIO_ITEMS = "subplot_owner_options_radio_items"

    RESAMPLING_FREQUENCY_DROPDOWN = "resampling_frequency_dropdown"

    REALIZATIONS_FILTER_OPTION = "realizations_filter_option"
    REALIZATIONS_FILTER_SELECTOR = "realizations_filter_selector"
    REALIZATIONS_FILTER_SLIDER = "realizations_filter_slider"
    REALIZATIONS_FILTER_SPAN = "realizations_filter_span"
    STATISTICS_FROM_RADIO_ITEMS = "statistics_from_radio_items"

    TOUR_STEP_MAIN_LAYOUT = "tour_step_main_layout"
    TOUR_STEP_SETTINGS_LAYOUT = "tour_step_settings_layout"
    TOUR_STEP_GROUP_BY = "tour_step_group_by"
    TOUR_STEP_DELTA_ENSEMBLE = "tour_step_delta_ensemble"
    TOUR_STEP_OPTIONS = "tour_step_options"


# pylint: disable = too-many-arguments
def main_layout(
    get_uuid: Callable,
    ensemble_names: List[str],
    vector_selector_data: list,
    vector_calculator_data: list,
    predefined_expressions: List[ExpressionInfo],
    realizations: List[int],
    disable_resampling_dropdown: bool,
    selected_resampling_frequency: Frequency,
    selected_visualization: VisualizationOptions,
    selected_vectors: Optional[List[str]] = None,
) -> html.Div:
    return wcc.FlexBox(
        id=get_uuid(LayoutElements.TOUR_STEP_MAIN_LAYOUT),
        children=[
            # Settings layout
            wcc.FlexColumn(
                id=get_uuid(LayoutElements.TOUR_STEP_SETTINGS_LAYOUT),
                children=wcc.Frame(
                    style={"height": "90vh"},
                    children=__settings_layout(
                        get_uuid=get_uuid,
                        ensembles=ensemble_names,
                        vector_selector_data=vector_selector_data,
                        vector_calculator_data=vector_calculator_data,
                        predefined_expressions=predefined_expressions,
                        realizations=realizations,
                        disable_resampling_dropdown=disable_resampling_dropdown,
                        selected_resampling_frequency=selected_resampling_frequency,
                        selected_visualization=selected_visualization,
                        selected_vectors=selected_vectors,
                    ),
                ),
            ),
            # Graph layout
            wcc.FlexColumn(
                flex=4,
                children=[
                    wcc.Frame(
                        style={"height": "90vh"},
                        highlight=False,
                        color="white",
                        children=[
                            wcc.Graph(
                                style={"height": "85vh"},
                                id=get_uuid(LayoutElements.GRAPH),
                            ),
                            dcc.Store(
                                # NOTE:Used to trigger graph update callback if data has
                                # changed, i.e. no change of regular INPUT html-elements
                                id=get_uuid(
                                    LayoutElements.GRAPH_DATA_HAS_CHANGED_TRIGGER
                                ),
                                data=0,
                            ),
                        ],
                    )
                ],
            ),
        ],
    )


# pylint: disable = too-many-arguments
def __settings_layout(
    get_uuid: Callable,
    ensembles: List[str],
    vector_selector_data: list,
    vector_calculator_data: list,
    predefined_expressions: List[ExpressionInfo],
    realizations: List[int],
    disable_resampling_dropdown: bool,
    selected_resampling_frequency: Frequency,
    selected_visualization: VisualizationOptions,
    selected_vectors: Optional[List[str]] = None,
) -> html.Div:
    return html.Div(
        children=[
            wcc.Selectors(
                label="Group By",
                id=get_uuid(LayoutElements.TOUR_STEP_GROUP_BY),
                children=[
                    wcc.RadioItems(
                        id=get_uuid(LayoutElements.SUBPLOT_OWNER_OPTIONS_RADIO_ITEMS),
                        options=[
                            {
                                "label": "Time Series",
                                "value": SubplotGroupByOptions.VECTOR.value,
                            },
                            {
                                "label": "Ensemble",
                                "value": SubplotGroupByOptions.ENSEMBLE.value,
                            },
                        ],
                        value=SubplotGroupByOptions.VECTOR.value,
                    ),
                ],
            ),
            wcc.Selectors(
                label="Resampling frequency",
                children=[
                    wcc.Dropdown(
                        id=get_uuid(LayoutElements.RESAMPLING_FREQUENCY_DROPDOWN),
                        clearable=False,
                        disabled=disable_resampling_dropdown,
                        options=[
                            {
                                "label": frequency.value,
                                "value": frequency.value,
                            }
                            for frequency in Frequency
                        ],
                        value=selected_resampling_frequency,
                    ),
                    wcc.Label(
                        "NB: Disabled for presampled data",
                        style={"font-style": "italic"}
                        if disable_resampling_dropdown
                        else {"display": "none"},
                    ),
                ],
            ),
            wcc.Selectors(
                label="Ensembles",
                children=[
                    wcc.Dropdown(
                        label="Selected ensembles",
                        id=get_uuid(LayoutElements.ENSEMBLES_DROPDOWN),
                        clearable=False,
                        multi=True,
                        options=[
                            {"label": ensemble, "value": ensemble}
                            for ensemble in ensembles
                        ],
                        value=None if len(ensembles) <= 0 else [ensembles[0]],
                    ),
                    wcc.Selectors(
                        label="Delta Ensembles",
                        id=get_uuid(LayoutElements.TOUR_STEP_DELTA_ENSEMBLE),
                        open_details=False,
                        children=[
                            __delta_ensemble_creator_layout(
                                get_uuid=get_uuid,
                                ensembles=ensembles,
                            )
                        ],
                    ),
                ],
            ),
            wcc.Selectors(
                label="Time Series",
                children=[
                    wsc.VectorSelector(
                        id=get_uuid(LayoutElements.VECTOR_SELECTOR),
                        maxNumSelectedNodes=3,
                        data=vector_selector_data,
                        persistence=True,
                        persistence_type="session",
                        selectedTags=[]
                        if selected_vectors is None
                        else selected_vectors,
                        numSecondsUntilSuggestionsAreShown=0.5,
                        lineBreakAfterTag=True,
                        customVectorDefinitions=get_custom_vector_definitions_from_expressions(
                            predefined_expressions
                        ),
                    ),
                    html.Button(
                        "Vector Calculator",
                        id=get_uuid(LayoutElements.VECTOR_CALCULATOR_OPEN_BUTTON),
                        style={
                            "margin-top": "5px",
                            "margin-bottom": "5px",
                        },
                    ),
                ],
            ),
            wcc.Selectors(
                label="Visualization",
                children=[
                    wcc.RadioItems(
                        id=get_uuid(LayoutElements.VISUALIZATION_RADIO_ITEMS),
                        options=[
                            {
                                "label": "Individual realizations",
                                "value": VisualizationOptions.REALIZATIONS.value,
                            },
                            {
                                "label": "Statistical lines",
                                "value": VisualizationOptions.STATISTICS.value,
                            },
                            {
                                "label": "Statistical fanchart",
                                "value": VisualizationOptions.FANCHART.value,
                            },
                            {
                                "label": "Statistics + Realizations",
                                "value": VisualizationOptions.STATISTICS_AND_REALIZATIONS,
                            },
                        ],
                        value=selected_visualization.value,
                    ),
                    wcc.Selectors(
                        label="Options",
                        id=get_uuid(LayoutElements.TOUR_STEP_OPTIONS),
                        children=__plot_options_layout(
                            get_uuid=get_uuid,
                            selected_visualization=selected_visualization,
                        ),
                    ),
                ],
            ),
            wcc.Selectors(
                label="Filter Realizations",
                children=__realization_filters(get_uuid, realizations),
            ),
            __vector_calculator_modal_layout(
                get_uuid=get_uuid,
                vector_data=vector_calculator_data,
                predefined_expressions=predefined_expressions,
            ),
            dcc.Store(
                id=get_uuid(LayoutElements.VECTOR_CALCULATOR_EXPRESSIONS),
                data=predefined_expressions,
            ),
            dcc.Store(
                id=get_uuid(LayoutElements.VECTOR_CALCULATOR_EXPRESSIONS_OPEN_MODAL),
                data=predefined_expressions,
            ),
        ]
    )


def __vector_calculator_modal_layout(
    get_uuid: Callable,
    vector_data: list,
    predefined_expressions: List[ExpressionInfo],
) -> dbc.Modal:
    return dbc.Modal(
        style={"marginTop": "20vh", "width": "1300px"},
        children=[
            dbc.ModalHeader("Vector Calculator"),
            dbc.ModalBody(
                html.Div(
                    children=[
                        wsc.VectorCalculator(
                            id=get_uuid(LayoutElements.VECTOR_CALCULATOR),
                            vectors=vector_data,
                            expressions=predefined_expressions,
                        )
                    ],
                ),
            ),
        ],
        id=get_uuid(LayoutElements.VECTOR_CALCULATOR_MODAL),
        size="lg",
        centered=True,
    )


def __delta_ensemble_creator_layout(
    get_uuid: Callable, ensembles: List[str]
) -> html.Div:
    return html.Div(
        children=[
            wcc.Dropdown(
                label="Ensemble A",
                id=get_uuid(LayoutElements.DELTA_ENSEMBLE_A_DROPDOWN),
                clearable=False,
                options=[{"label": i, "value": i} for i in ensembles],
                value=ensembles[0],
                style={"min-width": "60px"},
            ),
            wcc.Dropdown(
                label="Ensemble B",
                id=get_uuid(LayoutElements.DELTA_ENSEMBLE_B_DROPDOWN),
                clearable=False,
                options=[{"label": i, "value": i} for i in ensembles],
                value=ensembles[-1],
                style={"min-width": "60px"},
            ),
            html.Button(
                "Create",
                id=get_uuid(LayoutElements.DELTA_ENSEMBLE_CREATE_BUTTON),
                n_clicks=0,
                style={
                    "margin-top": "5px",
                    "margin-bottom": "5px",
                    "min-width": "20px",
                },
            ),
            __delta_ensemble_table_layout(get_uuid),
            dcc.Store(
                id=get_uuid(LayoutElements.CREATED_DELTA_ENSEMBLES),
                data=[],
            ),  # TODO: Add predefined deltas?
        ]
    )


def __delta_ensemble_table_layout(get_uuid: Callable) -> dash_table.DataTable:
    return dash_table.DataTable(
        id=get_uuid(LayoutElements.CREATED_DELTA_ENSEMBLE_NAMES_TABLE),
        columns=(
            [
                {
                    "id": get_uuid(
                        LayoutElements.CREATED_DELTA_ENSEMBLE_NAMES_TABLE_COLUMN
                    ),
                    "name": "Created Delta (A-B)",
                }
            ]
        ),
        data=[],
        fixed_rows={"headers": True},
        style_as_list_view=True,
        style_cell={"textAlign": "left"},
        style_header={"fontWeight": "bold"},
        style_table={
            "maxHeight": "150px",
            "overflowY": "auto",
        },
        editable=False,
    )


def __realization_filters(get_uuid: Callable, realizations: List[int]) -> html.Div:
    return html.Div(
        children=[
            html.Div(
                style={"display": "inline-flex"},
                children=[
                    html.Label(
                        "Realizations: ",
                        style={"font-weight": "bold"},
                    ),
                    html.Label(
                        id=get_uuid(LayoutElements.REALIZATIONS_FILTER_SPAN),
                        style={
                            "margin-left": "10px",
                            "margin-bottom": "5px",
                        },
                        children=f"{min(realizations)}-{max(realizations)}",
                    ),
                ],
            ),
            wcc.RadioItems(
                label="Statistics calculated from:",
                id=get_uuid(LayoutElements.STATISTICS_FROM_RADIO_ITEMS),
                style={"margin-bottom": "10px"},
                options=[
                    {
                        "label": "All",
                        "value": StatisticsFromOptions.ALL_REALIZATIONS.value,
                    },
                    {
                        "label": "Selected",
                        "value": StatisticsFromOptions.SELECTED_REALIZATIONS.value,
                    },
                ],
                value=StatisticsFromOptions.ALL_REALIZATIONS.value,
                vertical=False,
            ),
            html.Div(
                children=wcc.Select(
                    id=get_uuid(LayoutElements.REALIZATIONS_FILTER_SELECTOR),
                    options=[{"label": i, "value": i} for i in realizations],
                    value=realizations,
                    size=min(10, len(realizations)),
                ),
            ),
        ],
    )


def __plot_options_layout(
    get_uuid: Callable,
    selected_visualization: VisualizationOptions,
) -> html.Div:
    return html.Div(
        children=[
            wcc.Checklist(
                id=get_uuid(LayoutElements.PLOT_TRACE_OPTIONS_CHECKLIST),
                options=[
                    {"label": "History", "value": TraceOptions.HISTORY.value},
                    {
                        "label": "Observation",
                        "value": TraceOptions.OBSERVATIONS.value,
                    },
                ],
                value=[TraceOptions.HISTORY.value, TraceOptions.OBSERVATIONS.value],
            ),
            wcc.Checklist(
                id=get_uuid(LayoutElements.PLOT_STATISTICS_OPTIONS_CHECKLIST),
                style={"display": "block"}
                if selected_visualization
                in [
                    VisualizationOptions.STATISTICS,
                    VisualizationOptions.STATISTICS_AND_REALIZATIONS,
                ]
                else {"display": "none"},
                options=[
                    {"label": "Mean", "value": StatisticsOptions.MEAN.value},
                    {
                        "label": "P10 (high)",
                        "value": StatisticsOptions.P10.value,
                    },
                    {
                        "label": "P50 (median)",
                        "value": StatisticsOptions.P50.value,
                    },
                    {
                        "label": "P90 (low)",
                        "value": StatisticsOptions.P90.value,
                    },
                    {"label": "Maximum", "value": StatisticsOptions.MAX.value},
                    {"label": "Minimum", "value": StatisticsOptions.MIN.value},
                ],
                value=[
                    StatisticsOptions.MEAN.value,
                    StatisticsOptions.P10.value,
                    StatisticsOptions.P90.value,
                ],
            ),
            wcc.Checklist(
                id=get_uuid(LayoutElements.PLOT_FANCHART_OPTIONS_CHECKLIST),
                style={"display": "block"}
                if VisualizationOptions.FANCHART == selected_visualization
                else {"display": "none"},
                options=[
                    {
                        "label": FanchartOptions.MEAN.value,
                        "value": FanchartOptions.MEAN.value,
                    },
                    {
                        "label": FanchartOptions.P10_P90.value,
                        "value": FanchartOptions.P10_P90.value,
                    },
                    {
                        "label": FanchartOptions.MIN_MAX.value,
                        "value": FanchartOptions.MIN_MAX.value,
                    },
                ],
                value=[
                    FanchartOptions.MEAN.value,
                    FanchartOptions.P10_P90.value,
                    FanchartOptions.MIN_MAX.value,
                ],
            ),
        ],
    )



from typing import Callable

import webviz_core_components as wcc
from dash import html


def main_layout(get_uuid: Callable, ensembles: list) -> wcc.FlexBox:
    """Main layout"""
    return wcc.FlexBox(
        id=get_uuid("layout"),
        children=[
            wcc.FlexColumn(
                flex=1,
                children=[
                    wcc.Frame(
                        style={"height": "82vh"},
                        children=[
                            selections_layout(get_uuid, ensembles),
                            options_layout(get_uuid),
                            filters_layout(get_uuid),
                        ],
                    )
                ],
            ),
            wcc.FlexColumn(
                flex=5,
                children=[
                    wcc.Frame(
                        style={"height": "82vh"},
                        highlight=False,
                        color="white",
                        children=html.Div(
                            id=get_uuid("grouptree_wrapper"),
                        ),
                    )
                ],
            ),
        ],
    )


def filters_layout(get_uuid: Callable) -> wcc.Selectors:
    """The filters part of the menu"""
    filters_uuid = get_uuid("filters")
    return wcc.Selectors(
        id=get_uuid("filters_layout"),
        label="Filters",
        children=[
            wcc.SelectWithLabel(
                label="Prod/Inj/Other",
                id={"id": filters_uuid, "element": "prod_inj_other"},
                options=[
                    {"label": "Production", "value": "prod"},
                    {"label": "Injection", "value": "inj"},
                    {"label": "Other", "value": "other"},
                ],
                value=["prod", "inj", "other"],
                multi=True,
                size=3,
            )
        ],
    )


def options_layout(get_uuid: Callable) -> wcc.Selectors:
    """The options part of the menu"""
    options_uuid = get_uuid("options")
    return wcc.Selectors(
        id=get_uuid("options_layout"),
        label="Options",
        children=[
            html.Div(
                id={"id": options_uuid, "element": "statistical_options"},
                children=[
                    wcc.RadioItems(
                        id={"id": options_uuid, "element": "statistical_option"},
                        options=[
                            {"label": "Mean", "value": "mean"},
                            {"label": "P10 (high)", "value": "p10"},
                            {"label": "P50 (median)", "value": "p50"},
                            {"label": "P90 (low)", "value": "p90"},
                            {"label": "Maximum", "value": "max"},
                            {"label": "Minimum", "value": "min"},
                        ],
                    )
                ],
            ),
            html.Div(
                id={"id": options_uuid, "element": "single_real_options"},
                children=[
                    wcc.Dropdown(
                        label="Realization",
                        id={"id": options_uuid, "element": "realization"},
                        options=[],
                        value=None,
                        multi=False,
                    )
                ],
            ),
        ],
    )


def selections_layout(get_uuid: Callable, ensembles: list) -> wcc.Selectors:
    """Layout for the component input options"""
    controls_uuid = get_uuid("controls")
    return wcc.Selectors(
        id=get_uuid("selections_layout"),
        label="Controls",
        children=[
            wcc.Dropdown(
                label="Ensemble",
                id={"id": controls_uuid, "element": "ensemble"},
                options=[{"label": ens, "value": ens} for ens in ensembles],
                clearable=False,
                value=ensembles[0],
            ),
            wcc.RadioItems(
                label="Mean or realization",
                id={"id": controls_uuid, "element": "tree_mode"},
            ),
        ],
    )



from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import webviz_core_components as wcc
from dash import Dash, Input, Output, State
from dash.exceptions import PreventUpdate

from ...._figures import BarChart, ScatterPlot
from .._business_logic import RftPlotterDataModel, correlate
from .._figures._formation_figure import FormationFigure
from .._layout import LayoutElements


def paramresp_callbacks(
    app: Dash, get_uuid: Callable, datamodel: RftPlotterDataModel
) -> None:
    @app.callback(
        Output(get_uuid(LayoutElements.PARAMRESP_PARAM), "value"),
        Input(get_uuid(LayoutElements.PARAMRESP_CORR_BARCHART_FIGURE), "clickData"),
        State(get_uuid(LayoutElements.PARAMRESP_CORRTYPE), "value"),
        prevent_initial_call=True,
    )
    def _update_param_from_clickdata(
        corr_vector_clickdata: Union[None, dict],
        corrtype: str,
    ) -> str:
        """Update the selected parameter from clickdata"""
        if corr_vector_clickdata is None or corrtype == "param_vs_sim":
            raise PreventUpdate
        return corr_vector_clickdata.get("points", [{}])[0].get("y")

    @app.callback(
        Output(get_uuid(LayoutElements.PARAMRESP_WELL), "value"),
        Output(get_uuid(LayoutElements.PARAMRESP_DATE_DROPDOWN), "children"),
        Output(get_uuid(LayoutElements.PARAMRESP_ZONE_DROPDOWN), "children"),
        Input(get_uuid(LayoutElements.PARAMRESP_CORR_BARCHART_FIGURE), "clickData"),
        State(get_uuid(LayoutElements.PARAMRESP_CORRTYPE), "value"),
        State(get_uuid(LayoutElements.PARAMRESP_WELL), "value"),
        State(get_uuid(LayoutElements.PARAMRESP_DATE), "value"),
        State(get_uuid(LayoutElements.PARAMRESP_ZONE), "value"),
        prevent_initial_call=True,
    )
    def _update_selections_from_clickdata(
        corr_vector_clickdata: Union[None, dict],
        corrtype: str,
        well: str,
        date: str,
        zone: str,
    ) -> Tuple[str, wcc.Dropdown, wcc.Dropdown]:
        """Update well, date and zone from clickdata"""
        if corr_vector_clickdata is None or corrtype == "sim_vs_param":
            raise PreventUpdate

        clickdata = corr_vector_clickdata.get("points", [{}])[0].get("y")
        ls_clickdata = clickdata.split()

        dates_in_well, zones_in_well = datamodel.well_dates_and_zones(well)
        dates_dropdown = wcc.Dropdown(
            label="Date",
            id=get_uuid(LayoutElements.PARAMRESP_DATE),
            options=[{"label": date, "value": date} for date in dates_in_well],
            value=ls_clickdata[1],
            clearable=False,
        )
        zones_dropdown = wcc.Dropdown(
            label="Zone",
            id=get_uuid(LayoutElements.PARAMRESP_ZONE),
            options=[{"label": zone, "value": zone} for zone in zones_in_well],
            value=ls_clickdata[2],
            clearable=False,
        )

        return ls_clickdata[0], dates_dropdown, zones_dropdown

    @app.callback(
        Output(get_uuid(LayoutElements.PARAMRESP_DATE), "options"),
        Output(get_uuid(LayoutElements.PARAMRESP_DATE), "value"),
        Output(get_uuid(LayoutElements.PARAMRESP_ZONE), "options"),
        Output(get_uuid(LayoutElements.PARAMRESP_ZONE), "value"),
        Input(get_uuid(LayoutElements.PARAMRESP_WELL), "value"),
        State(get_uuid(LayoutElements.PARAMRESP_ZONE), "value"),
    )
    def _update_date_and_zone(
        well: str, zone_state: str
    ) -> Tuple[List[Dict[str, str]], str, List[Dict[str, str]], str]:
        """Update dates and zones when selecting well. If the current
        selected zone is also present in the new well it will be kept as value.
        """
        dates_in_well, zones_in_well = datamodel.well_dates_and_zones(well)
        return (
            [{"label": date, "value": date} for date in dates_in_well],
            dates_in_well[0],
            [{"label": zone, "value": zone} for zone in zones_in_well],
            zone_state if zone_state in zones_in_well else zones_in_well[0],
        )

    @app.callback(
        Output(get_uuid(LayoutElements.PARAMRESP_CORR_BARCHART), "children"),
        Output(get_uuid(LayoutElements.PARAMRESP_SCATTERPLOT), "children"),
        Output(get_uuid(LayoutElements.PARAMRESP_FORMATIONS), "children"),
        Input(get_uuid(LayoutElements.PARAMRESP_ENSEMBLE), "value"),
        Input(get_uuid(LayoutElements.PARAMRESP_WELL), "value"),
        Input(get_uuid(LayoutElements.PARAMRESP_DATE), "value"),
        Input(get_uuid(LayoutElements.PARAMRESP_ZONE), "value"),
        Input(get_uuid(LayoutElements.PARAMRESP_PARAM), "value"),
        Input(get_uuid(LayoutElements.PARAMRESP_CORRTYPE), "value"),
    )
    # pylint: disable=too-many-locals
    def _update_paramresp_graphs(
        ensemble: str,
        well: str,
        date: str,
        zone: str,
        param: Optional[str],
        corrtype: str,
    ) -> List[Optional[Any]]:
        """Main callback to update the graphs:

        * ranked correlations bar chart
        * response vs param scatter plot
        * formations chart RFT pressure vs depth, colored by parameter value
        """
        (
            df,
            obs,
            obs_err,
            ens_params,
            ens_rfts,
        ) = datamodel.create_rft_and_param_pivot_table(
            ensemble=ensemble,
            well=well,
            date=date,
            zone=zone,
            keep_all_rfts=(corrtype == "param_vs_sim"),
        )
        current_key = f"{well} {date} {zone}"

        if df is None:
            # This happens if the filtering criterias returns no data
            # Could f.ex happen when there are ensembles with different well names
            return ["No data matching the given filter criterias."] * 3
        if param is not None and param not in ens_params:
            # This happens if the selected parameter does not exist in the
            # selected ensemble
            return ["The selected parameter not valid for selected ensemble."] * 3
        if not ens_params:
            # This happens if there are multiple ensembles and one of the ensembles
            # doesn't have non-constant parameters.
            return ["The selected ensemble has no non-constant parameters."] * 3

        if corrtype == "sim_vs_param" or param is None:
            corrseries = correlate(df[ens_params + [current_key]], current_key)
            param = param if param is not None else corrseries.abs().idxmax()
            corr_title = f"{current_key} vs parameters"
            scatter_x, scatter_y, highlight_bar = param, current_key, param

        if corrtype == "param_vs_sim":
            corrseries = correlate(df[ens_rfts + [param]], param)
            corr_title = f"{param} vs simulated RFTs"
            scatter_x, scatter_y, highlight_bar = param, current_key, current_key

        # Correlation bar chart
        corrfig = BarChart(corrseries, n_rows=15, title=corr_title, orientation="h")
        corrfig.color_bars(highlight_bar, "#007079", 0.5)

        # Scatter plot
        scatterplot = ScatterPlot(
            df, scatter_y, scatter_x, "#007079", f"{current_key} vs {param}"
        )
        scatterplot.add_vertical_line_with_error(
            obs,
            obs_err,
            df[param].min(),
            df[param].max(),
        )

        # Formations plot
        formations_figure = FormationFigure(
            well=well,
            ertdf=datamodel.ertdatadf,
            enscolors=datamodel.enscolors,
            depth_option="TVD",
            date=date,
            ensembles=[ensemble],
            simdf=datamodel.simdf,
            obsdf=datamodel.obsdatadf,
        )

        if datamodel.formations is not None:
            formations_figure.add_formation(datamodel.formationdf, fill_color=False)

        formations_figure.add_simulated_lines("realization")
        formations_figure.add_additional_observations()
        formations_figure.add_ert_observed()

        df_value_norm = datamodel.get_param_real_and_value_df(
            ensemble, parameter=param, normalize=True
        )
        formations_figure.color_by_param_value(df_value_norm, param)

        return [
            wcc.Graph(
                style={"height": "40vh"},
                config={"displayModeBar": False},
                figure=corrfig.figure,
                id=get_uuid(LayoutElements.PARAMRESP_CORR_BARCHART_FIGURE),
            ),
            wcc.Graph(
                style={"height": "40vh"},
                config={"displayModeBar": False},
                figure=scatterplot.figure,
            ),
            wcc.Graph(
                style={"height": "87vh"},
                figure=formations_figure.figure,
            ),
        ]



import copy
from typing import Callable, Dict, List, Optional, Tuple, Union

import dash
import pandas as pd
import webviz_subsurface_components as wsc
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate
from webviz_config import EncodedFile, WebvizPluginABC
from webviz_config._theme_class import WebvizConfigTheme
from webviz_subsurface_components import ExpressionInfo, ExternalParseData

from webviz_subsurface._providers import Frequency
from webviz_subsurface._utils.formatting import printable_int_list
from webviz_subsurface._utils.unique_theming import unique_colors
from webviz_subsurface._utils.vector_calculator import (
    add_expressions_to_vector_selector_data,
    get_custom_vector_definitions_from_expressions,
    get_selected_expressions,
)
from webviz_subsurface._utils.vector_selector import (
    is_vector_name_in_vector_selector_data,
)

from ._layout import LayoutElements
from ._property_serialization import (
    EnsembleSubplotBuilder,
    GraphFigureBuilderBase,
    VectorSubplotBuilder,
)
from .types import (
    DeltaEnsemble,
    DerivedVectorsAccessor,
    FanchartOptions,
    ProviderSet,
    StatisticsFromOptions,
    StatisticsOptions,
    SubplotGroupByOptions,
    TraceOptions,
    VisualizationOptions,
)
from .utils.delta_ensemble_utils import create_delta_ensemble_names
from .utils.derived_ensemble_vectors_accessor_utils import (
    create_derived_vectors_accessor_dict,
)
from .utils.from_timeseries_cumulatives import datetime_to_intervalstr
from .utils.history_vectors import create_history_vectors_df
from .utils.provider_set_utils import create_vector_plot_titles_from_provider_set
from .utils.trace_line_shape import get_simulation_line_shape
from .utils.vector_statistics import create_vectors_statistics_df


# pylint: disable = too-many-arguments, too-many-branches, too-many-locals, too-many-statements
def plugin_callbacks(
    app: dash.Dash,
    get_uuid: Callable,
    get_data_output: Output,
    get_data_requested: Input,
    input_provider_set: ProviderSet,
    theme: WebvizConfigTheme,
    initial_selected_vectors: List[str],
    vector_selector_base_data: list,
    observations: dict,  # TODO: Improve typehint?
    line_shape_fallback: str = "linear",
) -> None:
    # TODO: Consider adding: presampled_frequency: Optional[Frequency] argument for use when
    # providers are presampled. To keep track of sampling frequency, and not depend on dropdown
    # value for ViewElements.RESAMPLING_FREQUENCY_DROPDOWN (dropdown disabled when providers are
    # presampled)
    @app.callback(
        Output(get_uuid(LayoutElements.GRAPH), "figure"),
        [
            Input(
                get_uuid(LayoutElements.VECTOR_SELECTOR),
                "selectedNodes",
            ),
            Input(get_uuid(LayoutElements.ENSEMBLES_DROPDOWN), "value"),
            Input(
                get_uuid(LayoutElements.VISUALIZATION_RADIO_ITEMS),
                "value",
            ),
            Input(
                get_uuid(LayoutElements.PLOT_STATISTICS_OPTIONS_CHECKLIST),
                "value",
            ),
            Input(
                get_uuid(LayoutElements.PLOT_FANCHART_OPTIONS_CHECKLIST),
                "value",
            ),
            Input(
                get_uuid(LayoutElements.PLOT_TRACE_OPTIONS_CHECKLIST),
                "value",
            ),
            Input(
                get_uuid(LayoutElements.SUBPLOT_OWNER_OPTIONS_RADIO_ITEMS),
                "value",
            ),
            Input(
                get_uuid(LayoutElements.RESAMPLING_FREQUENCY_DROPDOWN),
                "value",
            ),
            Input(get_uuid(LayoutElements.REALIZATIONS_FILTER_SELECTOR), "value"),
            Input(
                get_uuid(LayoutElements.STATISTICS_FROM_RADIO_ITEMS),
                "value",
            ),
            Input(
                get_uuid(LayoutElements.GRAPH_DATA_HAS_CHANGED_TRIGGER),
                "data",
            ),
        ],
        [
            State(
                get_uuid(LayoutElements.CREATED_DELTA_ENSEMBLES),
                "data",
            ),
            State(
                get_uuid(LayoutElements.VECTOR_CALCULATOR_EXPRESSIONS),
                "data",
            ),
            State(get_uuid(LayoutElements.ENSEMBLES_DROPDOWN), "options"),
        ],
    )
    def _update_graph(
        vectors: List[str],
        selected_ensembles: List[str],
        visualization_value: str,
        statistics_option_values: List[str],
        fanchart_option_values: List[str],
        trace_option_values: List[str],
        subplot_owner_options_value: str,
        resampling_frequency_value: str,
        selected_realizations: List[int],
        statistics_calculated_from_value: str,
        __graph_data_has_changed_trigger: int,
        delta_ensembles: List[DeltaEnsemble],
        vector_calculator_expressions: List[ExpressionInfo],
        ensemble_dropdown_options: List[dict],
    ) -> dict:
        """Callback to update all graphs based on selections

        * De-serialize from JSON serializable format to strongly typed and filtered format
        * Business logic:
            * Functionality with "strongly typed" and filtered input format - functions and
            classes
            * ProviderSet for EnsembleSummaryProviders, i.e. input_provider_set
            * DerivedEnsembleVectorsAccessor to access derived vector data from ensembles
            with single providers or delta ensemble with two providers
            * GraphFigureBuilder to create graph with subplots per vector or subplots per
            ensemble, using VectorSubplotBuilder and EnsembleSubplotBuilder, respectively
        * Create/build property serialization in FigureBuilder by use of business logic data

        NOTE: __graph_data_has_changed_trigger is only used to trigger callback when change of
        graphs data has changed and re-render of graph is necessary. E.g. when a selected expression
        from the VectorCalculatorgets edited, without changing the expression name - i.e.
        VectorSelector selectedNodes remain unchanged.
        """
        if not isinstance(selected_ensembles, list):
            raise TypeError("ensembles should always be of type list")

        if vectors is None:
            vectors = initial_selected_vectors

        # Retrieve the selected expressions
        selected_expressions = get_selected_expressions(
            vector_calculator_expressions, vectors
        )

        # Convert from string values to enum types
        visualization = VisualizationOptions(visualization_value)
        statistics_options = [
            StatisticsOptions(elm) for elm in statistics_option_values
        ]
        fanchart_options = [FanchartOptions(elm) for elm in fanchart_option_values]
        trace_options = [TraceOptions(elm) for elm in trace_option_values]
        subplot_owner = SubplotGroupByOptions(subplot_owner_options_value)
        resampling_frequency = Frequency.from_string_value(resampling_frequency_value)
        all_ensemble_names = [option["value"] for option in ensemble_dropdown_options]
        statistics_from_option = StatisticsFromOptions(statistics_calculated_from_value)

        # Prevent update if realization filtering is not affecting pure statistics plot
        # TODO: Refactor code or create utility for getting trigger ID in a "cleaner" way?
        ctx = dash.callback_context.triggered
        trigger_id = ctx[0]["prop_id"].split(".")[0]
        if (
            trigger_id == get_uuid(LayoutElements.REALIZATIONS_FILTER_SELECTOR)
            and statistics_from_option is StatisticsFromOptions.ALL_REALIZATIONS
            and visualization
            in [
                VisualizationOptions.STATISTICS,
                VisualizationOptions.FANCHART,
            ]
        ):
            raise PreventUpdate

        # Create dict of derived vectors accessors for selected ensembles
        derived_vectors_accessors: Dict[
            str, DerivedVectorsAccessor
        ] = create_derived_vectors_accessor_dict(
            ensembles=selected_ensembles,
            vectors=vectors,
            provider_set=input_provider_set,
            expressions=selected_expressions,
            delta_ensembles=delta_ensembles,
            resampling_frequency=resampling_frequency,
        )

        # TODO: How to get metadata for calculated vector?
        vector_line_shapes: Dict[str, str] = {
            vector: get_simulation_line_shape(
                line_shape_fallback,
                vector,
                input_provider_set.vector_metadata(vector),
            )
            for vector in vectors
        }

        figure_builder: GraphFigureBuilderBase
        if subplot_owner is SubplotGroupByOptions.VECTOR:
            # Create unique colors based on all ensemble names to preserve consistent colors
            ensemble_colors = unique_colors(all_ensemble_names, theme)
            vector_titles = create_vector_plot_titles_from_provider_set(
                vectors, selected_expressions, input_provider_set
            )
            figure_builder = VectorSubplotBuilder(
                vectors,
                vector_titles,
                ensemble_colors,
                resampling_frequency,
                vector_line_shapes,
                theme,
            )
        elif subplot_owner is SubplotGroupByOptions.ENSEMBLE:
            vector_colors = unique_colors(vectors, theme)
            figure_builder = EnsembleSubplotBuilder(
                vectors,
                selected_ensembles,
                vector_colors,
                resampling_frequency,
                vector_line_shapes,
                theme,
            )
        else:
            raise PreventUpdate

        # Get all realizations if statistics accross all realizations are requested
        is_statistics_from_all_realizations = (
            statistics_from_option == StatisticsFromOptions.ALL_REALIZATIONS
            and visualization
            in [
                VisualizationOptions.FANCHART,
                VisualizationOptions.STATISTICS,
                VisualizationOptions.STATISTICS_AND_REALIZATIONS,
            ]
        )

        # Plotting per derived vectors accessor
        for ensemble, accessor in derived_vectors_accessors.items():
            # Realization query - realizations query for accessor
            # - Get non-filter query, None, if statistics from all realizations is needed
            # - Create valid realizations query for accessor otherwise:
            #   * List[int]: Filtered valid realizations, empty list if none are valid
            #   * None: Get all realizations, i.e. non-filtered query
            realizations_query = (
                None
                if is_statistics_from_all_realizations
                else accessor.create_valid_realizations_query(selected_realizations)
            )

            # If all selected realizations are invalid for accessor - empty list
            if realizations_query == []:
                continue

            # TODO: Consider to remove list vectors_df_list and use pd.concat to obtain
            # one single dataframe with vector columns. NB: Assumes equal sampling rate
            # for each vector type - i.e equal number of rows in dataframes

            # Retrive vectors data from accessor
            vectors_df_list: List[pd.DataFrame] = []
            if accessor.has_provider_vectors():
                vectors_df_list.append(
                    accessor.get_provider_vectors_df(realizations=realizations_query)
                )
            if accessor.has_interval_and_average_vectors():
                vectors_df_list.append(
                    accessor.create_interval_and_average_vectors_df(
                        realizations=realizations_query
                    )
                )
            if accessor.has_vector_calculator_expressions():
                vectors_df_list.append(
                    accessor.create_calculated_vectors_df(
                        realizations=realizations_query
                    )
                )

            for vectors_df in vectors_df_list:
                if visualization == VisualizationOptions.REALIZATIONS:
                    # Show selected realizations - only filter df if realizations filter
                    # query is not performed
                    figure_builder.add_realizations_traces(
                        vectors_df
                        if realizations_query
                        else vectors_df[vectors_df["REAL"].isin(selected_realizations)],
                        ensemble,
                    )
                if visualization == VisualizationOptions.STATISTICS:
                    vectors_statistics_df = create_vectors_statistics_df(vectors_df)
                    figure_builder.add_statistics_traces(
                        vectors_statistics_df,
                        ensemble,
                        statistics_options,
                    )
                if visualization == VisualizationOptions.FANCHART:
                    vectors_statistics_df = create_vectors_statistics_df(vectors_df)
                    figure_builder.add_fanchart_traces(
                        vectors_statistics_df,
                        ensemble,
                        fanchart_options,
                    )
                if visualization == VisualizationOptions.STATISTICS_AND_REALIZATIONS:
                    # Configure line width and color scaling to easier separate
                    # statistics traces and realization traces.
                    # Show selected realizations - only filter df if realizations filter
                    # query is not performed
                    figure_builder.add_realizations_traces(
                        vectors_df
                        if realizations_query
                        else vectors_df[vectors_df["REAL"].isin(selected_realizations)],
                        ensemble,
                        color_lightness_scale=150.0,
                    )
                    # Add statistics on top
                    vectors_statistics_df = create_vectors_statistics_df(vectors_df)
                    figure_builder.add_statistics_traces(
                        vectors_statistics_df,
                        ensemble,
                        statistics_options,
                        line_width=3,
                    )

        # Retrieve selected input providers
        selected_input_providers = ProviderSet(
            {
                name: provider
                for name, provider in input_provider_set.items()
                if name in selected_ensembles
            }
        )

        # Do not add observations if only delta ensembles are selected
        is_only_delta_ensembles = (
            len(selected_input_providers.names()) == 0
            and len(derived_vectors_accessors) > 0
        )
        if (
            observations
            and TraceOptions.OBSERVATIONS in trace_options
            and not is_only_delta_ensembles
        ):
            for vector in vectors:
                vector_observations = observations.get(vector)
                if vector_observations:
                    figure_builder.add_vector_observations(vector, vector_observations)

        # Add history trace
        # TODO: Improve when new history vector input format is in place
        if TraceOptions.HISTORY in trace_options:
            if (
                isinstance(figure_builder, VectorSubplotBuilder)
                and len(selected_input_providers.names()) > 0
            ):
                # Add history trace using first selected ensemble
                name = selected_input_providers.names()[0]
                provider = selected_input_providers.provider(name)
                vector_names = provider.vector_names()

                provider_vectors = [elm for elm in vectors if elm in vector_names]
                if provider_vectors:
                    history_vectors_df = create_history_vectors_df(
                        provider, provider_vectors, resampling_frequency
                    )
                    # TODO: Handle check of non-empty dataframe better?
                    if (
                        not history_vectors_df.empty
                        and "DATE" in history_vectors_df.columns
                    ):
                        figure_builder.add_history_traces(history_vectors_df)

            if isinstance(figure_builder, EnsembleSubplotBuilder):
                # Add history trace for each ensemble
                for name, provider in selected_input_providers.items():
                    vector_names = provider.vector_names()

                    provider_vectors = [elm for elm in vectors if elm in vector_names]
                    if provider_vectors:
                        history_vectors_df = create_history_vectors_df(
                            provider, provider_vectors, resampling_frequency
                        )
                        # TODO: Handle check of non-empty dataframe better?
                        if (
                            not history_vectors_df.empty
                            and "DATE" in history_vectors_df.columns
                        ):
                            figure_builder.add_history_traces(
                                history_vectors_df,
                                name,
                            )

        # Create legends when all data is added to figure
        figure_builder.create_graph_legends()

        return figure_builder.get_serialized_figure()

    @app.callback(
        get_data_output,
        [get_data_requested],
        [
            State(
                get_uuid(LayoutElements.VECTOR_SELECTOR),
                "selectedNodes",
            ),
            State(get_uuid(LayoutElements.ENSEMBLES_DROPDOWN), "value"),
            State(
                get_uuid(LayoutElements.VISUALIZATION_RADIO_ITEMS),
                "value",
            ),
            State(
                get_uuid(LayoutElements.RESAMPLING_FREQUENCY_DROPDOWN),
                "value",
            ),
            State(get_uuid(LayoutElements.REALIZATIONS_FILTER_SELECTOR), "value"),
            State(
                get_uuid(LayoutElements.STATISTICS_FROM_RADIO_ITEMS),
                "value",
            ),
            State(
                get_uuid(LayoutElements.CREATED_DELTA_ENSEMBLES),
                "data",
            ),
            State(
                get_uuid(LayoutElements.VECTOR_CALCULATOR_EXPRESSIONS),
                "data",
            ),
        ],
    )
    def _user_download_data(
        data_requested: Union[int, None],
        vectors: List[str],
        selected_ensembles: List[str],
        visualization_value: str,
        resampling_frequency_value: str,
        selected_realizations: List[int],
        statistics_calculated_from_value: str,
        delta_ensembles: List[DeltaEnsemble],
        vector_calculator_expressions: List[ExpressionInfo],
    ) -> Union[EncodedFile, str]:
        """Callback to download data based on selections

        Retrieve vector data based on selected visualizations and filtered realizations

        NOTE:
        * Does not group based on "Group By" - data is stored per vector
        * All statistics included - no filtering on statistics selections
        * No history vector
        * No observation data
        """
        if data_requested is None:
            raise PreventUpdate

        if not isinstance(selected_ensembles, list):
            raise TypeError("ensembles should always be of type list")

        if vectors is None:
            vectors = initial_selected_vectors

        # Retrieve the selected expressions
        selected_expressions = get_selected_expressions(
            vector_calculator_expressions, vectors
        )

        # Convert from string values to enum types
        visualization = VisualizationOptions(visualization_value)
        resampling_frequency = Frequency.from_string_value(resampling_frequency_value)
        statistics_from_option = StatisticsFromOptions(statistics_calculated_from_value)

        # Create dict of derived vectors accessors for selected ensembles
        derived_vectors_accessors: Dict[
            str, DerivedVectorsAccessor
        ] = create_derived_vectors_accessor_dict(
            ensembles=selected_ensembles,
            vectors=vectors,
            provider_set=input_provider_set,
            expressions=selected_expressions,
            delta_ensembles=delta_ensembles,
            resampling_frequency=resampling_frequency,
        )

        # Dict with vector name as key and dataframe data as value
        vector_dataframe_dict: Dict[str, pd.DataFrame] = {}

        # Get all realizations if statistics accross all realizations are requested
        is_statistics_from_all_realizations = (
            statistics_from_option == StatisticsFromOptions.ALL_REALIZATIONS
            and visualization
            in [
                VisualizationOptions.FANCHART,
                VisualizationOptions.STATISTICS,
                VisualizationOptions.STATISTICS_AND_REALIZATIONS,
            ]
        )

        # Plotting per derived vectors accessor
        for ensemble, accessor in derived_vectors_accessors.items():
            # Realization query - realizations query for accessor
            # - Get non-filter query, None, if statistics from all realizations is needed
            # - Create valid realizations query for accessor otherwise:
            #   * List[int]: Filtered valid realizations, empty list if none are valid
            #   * None: Get all realizations, i.e. non-filtered query
            realizations_query = (
                None
                if is_statistics_from_all_realizations
                else accessor.create_valid_realizations_query(selected_realizations)
            )

            # If all selected realizations are invalid for accessor - empty list
            if realizations_query == []:
                continue

            # Retrive vectors data from accessor
            vectors_df_list: List[pd.DataFrame] = []
            if accessor.has_provider_vectors():
                vectors_df_list.append(
                    accessor.get_provider_vectors_df(realizations=realizations_query)
                )
            if accessor.has_interval_and_average_vectors():
                vectors_df_list.append(
                    accessor.create_interval_and_average_vectors_df(
                        realizations=realizations_query
                    )
                )
            if accessor.has_vector_calculator_expressions():
                vectors_df_list.append(
                    accessor.create_calculated_vectors_df(
                        realizations=realizations_query
                    )
                )

            # Append data for each vector
            for vectors_df in vectors_df_list:
                vector_names = [
                    elm for elm in vectors_df.columns if elm not in ["DATE", "REAL"]
                ]

                if visualization in [
                    VisualizationOptions.REALIZATIONS,
                    VisualizationOptions.STATISTICS_AND_REALIZATIONS,
                ]:
                    # NOTE: Should in theory not have situation with query of all realizations
                    # if not wanted
                    vectors_df_filtered = (
                        vectors_df
                        if realizations_query
                        else vectors_df[vectors_df["REAL"].isin(selected_realizations)]
                    )
                    for vector in vector_names:
                        vector_df = vectors_df_filtered[["DATE", "REAL", vector]]
                        row_count = vector_df.shape[0]
                        ensemble_name_list = [ensemble] * row_count
                        vector_df.insert(
                            loc=0, column="ENSEMBLE", value=ensemble_name_list
                        )

                        if vector.startswith(("AVG_", "INTVL_")):
                            vector_df["DATE"] = vector_df["DATE"].apply(
                                datetime_to_intervalstr, freq=resampling_frequency
                            )

                        vector_key = vector + "_realizations"
                        if vector_dataframe_dict.get(vector_key) is None:
                            vector_dataframe_dict[vector_key] = vector_df
                        else:
                            vector_dataframe_dict[vector_key] = pd.concat(
                                [vector_dataframe_dict[vector_key], vector_df],
                                ignore_index=True,
                                axis=0,
                            )

                if visualization in [
                    VisualizationOptions.STATISTICS,
                    VisualizationOptions.FANCHART,
                    VisualizationOptions.STATISTICS_AND_REALIZATIONS,
                ]:
                    vectors_statistics_df = create_vectors_statistics_df(vectors_df)

                    for vector in vector_names:
                        vector_statistics_df = vectors_statistics_df[["DATE", vector]]
                        row_count = vector_statistics_df.shape[0]
                        ensemble_name_list = [ensemble] * row_count
                        vector_statistics_df.insert(
                            loc=0, column="ENSEMBLE", value=ensemble_name_list
                        )

                        vector_key = vector + "_statistics"

                        if vector.startswith(("AVG_", "INTVL_")):
                            vector_statistics_df.loc[
                                :, ("DATE", "")
                            ] = vector_statistics_df.loc[:, ("DATE", "")].apply(
                                datetime_to_intervalstr, freq=resampling_frequency
                            )
                        if vector_dataframe_dict.get(vector_key) is None:
                            vector_dataframe_dict[vector_key] = vector_statistics_df
                        else:
                            vector_dataframe_dict[vector_key] = pd.concat(
                                [
                                    vector_dataframe_dict[vector_key],
                                    vector_statistics_df,
                                ],
                                ignore_index=True,
                                axis=0,
                            )

        # : is replaced with _ in filenames to stay within POSIX portable pathnames
        # (e.g. : is not valid in a Windows path)
        return WebvizPluginABC.plugin_data_compress(
            [
                {
                    "filename": f"{vector.replace(':', '_')}.csv",
                    "content": df.to_csv(index=False),
                }
                for vector, df in vector_dataframe_dict.items()
            ]
        )

    @app.callback(
        [
            Output(
                get_uuid(LayoutElements.PLOT_STATISTICS_OPTIONS_CHECKLIST),
                "style",
            ),
            Output(
                get_uuid(LayoutElements.PLOT_FANCHART_OPTIONS_CHECKLIST),
                "style",
            ),
        ],
        [
            Input(
                get_uuid(LayoutElements.VISUALIZATION_RADIO_ITEMS),
                "value",
            )
        ],
    )
    def _update_statistics_options_layout(selected_visualization: str) -> List[dict]:
        """Only show statistics checklist if in statistics mode"""

        # Convert to enum type
        selected_visualization = VisualizationOptions(selected_visualization)

        def get_style(visualization_options: List[VisualizationOptions]) -> dict:
            return (
                {"display": "block"}
                if selected_visualization in visualization_options
                else {"display": "none"}
            )

        statistics_options_style = get_style(
            [
                VisualizationOptions.STATISTICS,
                VisualizationOptions.STATISTICS_AND_REALIZATIONS,
            ]
        )
        fanchart_options_style = get_style([VisualizationOptions.FANCHART])

        return [statistics_options_style, fanchart_options_style]

    @app.callback(
        [
            Output(
                get_uuid(LayoutElements.CREATED_DELTA_ENSEMBLES),
                "data",
            ),
            Output(
                get_uuid(LayoutElements.CREATED_DELTA_ENSEMBLE_NAMES_TABLE),
                "data",
            ),
            Output(
                get_uuid(LayoutElements.ENSEMBLES_DROPDOWN),
                "options",
            ),
        ],
        [
            Input(
                get_uuid(LayoutElements.DELTA_ENSEMBLE_CREATE_BUTTON),
                "n_clicks",
            )
        ],
        [
            State(
                get_uuid(LayoutElements.CREATED_DELTA_ENSEMBLES),
                "data",
            ),
            State(
                get_uuid(LayoutElements.DELTA_ENSEMBLE_A_DROPDOWN),
                "value",
            ),
            State(
                get_uuid(LayoutElements.DELTA_ENSEMBLE_B_DROPDOWN),
                "value",
            ),
        ],
    )
    def _update_created_delta_ensembles_names(
        n_clicks: int,
        existing_delta_ensembles: List[DeltaEnsemble],
        ensemble_a: str,
        ensemble_b: str,
    ) -> Tuple[List[DeltaEnsemble], List[Dict[str, str]], List[Dict[str, str]]]:
        if n_clicks is None or n_clicks <= 0:
            raise PreventUpdate

        delta_ensemble = DeltaEnsemble(ensemble_a=ensemble_a, ensemble_b=ensemble_b)
        if delta_ensemble in existing_delta_ensembles:
            raise PreventUpdate

        new_delta_ensembles = existing_delta_ensembles
        new_delta_ensembles.append(delta_ensemble)

        # Create delta ensemble names
        new_delta_ensemble_names = create_delta_ensemble_names(new_delta_ensembles)

        table_data = _create_delta_ensemble_table_column_data(
            get_uuid(LayoutElements.CREATED_DELTA_ENSEMBLE_NAMES_TABLE_COLUMN),
            new_delta_ensemble_names,
        )

        ensemble_options = [
            {"label": ensemble, "value": ensemble}
            for ensemble in input_provider_set.names()
        ]
        for elm in new_delta_ensemble_names:
            ensemble_options.append({"label": elm, "value": elm})

        return (new_delta_ensembles, table_data, ensemble_options)

    @app.callback(
        Output(get_uuid(LayoutElements.VECTOR_CALCULATOR_MODAL), "is_open"),
        [
            Input(get_uuid(LayoutElements.VECTOR_CALCULATOR_OPEN_BUTTON), "n_clicks"),
        ],
        [State(get_uuid(LayoutElements.VECTOR_CALCULATOR_MODAL), "is_open")],
    )
    def _toggle_vector_calculator_modal(n_open_clicks: int, is_open: bool) -> bool:
        if n_open_clicks:
            return not is_open
        raise PreventUpdate

    @app.callback(
        Output(get_uuid(LayoutElements.VECTOR_CALCULATOR), "externalParseData"),
        Input(get_uuid(LayoutElements.VECTOR_CALCULATOR), "externalParseExpression"),
    )
    def _parse_vector_calculator_expression(
        expression: ExpressionInfo,
    ) -> ExternalParseData:
        if expression is None:
            raise PreventUpdate
        return wsc.VectorCalculator.external_parse_data(expression)

    @app.callback(
        [
            Output(
                get_uuid(LayoutElements.VECTOR_CALCULATOR_EXPRESSIONS),
                "data",
            ),
            Output(get_uuid(LayoutElements.VECTOR_SELECTOR), "data"),
            Output(get_uuid(LayoutElements.VECTOR_SELECTOR), "selectedTags"),
            Output(get_uuid(LayoutElements.VECTOR_SELECTOR), "customVectorDefinitions"),
            Output(
                get_uuid(LayoutElements.GRAPH_DATA_HAS_CHANGED_TRIGGER),
                "data",
            ),
        ],
        Input(get_uuid(LayoutElements.VECTOR_CALCULATOR_MODAL), "is_open"),
        [
            State(
                get_uuid(LayoutElements.VECTOR_CALCULATOR_EXPRESSIONS_OPEN_MODAL),
                "data",
            ),
            State(
                get_uuid(LayoutElements.VECTOR_CALCULATOR_EXPRESSIONS),
                "data",
            ),
            State(get_uuid(LayoutElements.VECTOR_SELECTOR), "selectedNodes"),
            State(get_uuid(LayoutElements.VECTOR_SELECTOR), "customVectorDefinitions"),
            State(
                get_uuid(LayoutElements.GRAPH_DATA_HAS_CHANGED_TRIGGER),
                "data",
            ),
        ],
    )
    def _update_vector_calculator_expressions_on_modal_close(
        is_modal_open: bool,
        new_expressions: List[ExpressionInfo],
        current_expressions: List[ExpressionInfo],
        current_selected_vectors: List[str],
        current_custom_vector_definitions: dict,
        graph_data_has_changed_counter: int,
    ) -> list:
        """Update vector calculator expressions, propagate expressions to VectorSelectors,
        update current selections and trigger re-rendering of graphing if necessary
        """
        if is_modal_open or (new_expressions == current_expressions):
            raise PreventUpdate

        # Create current selected expressions for comparison - Deep copy!
        current_selected_expressions = get_selected_expressions(
            current_expressions, current_selected_vectors
        )

        # Create new vector selector data - Deep copy!
        new_vector_selector_data = copy.deepcopy(vector_selector_base_data)
        add_expressions_to_vector_selector_data(
            new_vector_selector_data, new_expressions
        )

        # Create new selected vectors - from new expressions
        new_selected_vectors = _create_new_selected_vectors(
            current_selected_vectors,
            current_expressions,
            new_expressions,
            new_vector_selector_data,
        )

        # Get new selected expressions
        new_selected_expressions = get_selected_expressions(
            new_expressions, new_selected_vectors
        )

        # Get new custom vector definitions
        new_custom_vector_definitions = get_custom_vector_definitions_from_expressions(
            new_expressions
        )

        # Prevent updates if unchanged
        if new_custom_vector_definitions == current_custom_vector_definitions:
            new_custom_vector_definitions = dash.no_update

        if new_selected_vectors == current_selected_vectors:
            new_selected_vectors = dash.no_update

        # If selected expressions are edited - Only trigger graph data update property when needed,
        # i.e. names are unchanged and selectedNodes for VectorSelector remains unchanged.
        new_graph_data_has_changed_counter = dash.no_update
        if (
            new_selected_expressions != current_selected_expressions
            and new_selected_vectors == dash.no_update
        ):
            new_graph_data_has_changed_counter = graph_data_has_changed_counter + 1

        return [
            new_expressions,
            new_vector_selector_data,
            new_selected_vectors,
            new_custom_vector_definitions,
            new_graph_data_has_changed_counter,
        ]

    @app.callback(
        Output(
            get_uuid(LayoutElements.VECTOR_CALCULATOR_EXPRESSIONS_OPEN_MODAL),
            "data",
        ),
        Input(get_uuid(LayoutElements.VECTOR_CALCULATOR), "expressions"),
    )
    def _update_vector_calculator_expressions_when_modal_open(
        expressions: List[ExpressionInfo],
    ) -> list:
        new_expressions: List[ExpressionInfo] = [
            elm for elm in expressions if elm["isValid"]
        ]
        return new_expressions

    @app.callback(
        Output(get_uuid(LayoutElements.REALIZATIONS_FILTER_SPAN), "children"),
        Input(get_uuid(LayoutElements.REALIZATIONS_FILTER_SELECTOR), "value"),
    )
    def _update_realization_range(realizations: List[int]) -> Optional[str]:
        if not realizations:
            raise PreventUpdate

        realizations_filter_text = printable_int_list(realizations)

        return realizations_filter_text


def _create_delta_ensemble_table_column_data(
    column_name: str, ensemble_names: List[str]
) -> List[Dict[str, str]]:
    return [{column_name: elm} for elm in ensemble_names]


def _create_new_selected_vectors(
    existing_selected_vectors: List[str],
    existing_expressions: List[ExpressionInfo],
    new_expressions: List[ExpressionInfo],
    new_vector_selector_data: list,
) -> List[str]:
    valid_selections: List[str] = []
    for vector in existing_selected_vectors:
        new_vector: Optional[str] = vector

        # Get id if vector is among existing expressions
        dropdown_id = next(
            (elm["id"] for elm in existing_expressions if elm["name"] == vector),
            None,
        )
        # Find id among new expressions to get new/edited name
        if dropdown_id:
            new_vector = next(
                (elm["name"] for elm in new_expressions if elm["id"] == dropdown_id),
                None,
            )

        # Append if vector name exist among data
        if new_vector is not None and is_vector_name_in_vector_selector_data(
            new_vector, new_vector_selector_data
        ):
            valid_selections.append(new_vector)
    return valid_selections
